# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

# version of this python module
__version__ = "1.2.10"
# matching version of the SigMF specification
__specification__ = "1.2.5"

from . import archive, archivereader, error, schema, sigmffile, utils, validate
from .archive import SigMFArchive
from .archivereader import SigMFArchiveReader
from .sigmffile import SigMFCollection, SigMFFile



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Defines SigMF exception classes."""


class SigMFError(Exception):
    """SigMF base exception."""


class SigMFValidationError(SigMFError):
    """Exceptions related to validating SigMF metadata."""


class SigMFAccessError(SigMFError):
    """Exceptions related to accessing the contents of SigMF metadata, notably
    when expected fields are missing or accessing out of bounds captures."""


class SigMFFileError(SigMFError):
    """Exceptions related to reading or writing SigMF files or archives."""



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Create and extract SigMF archives."""

import io
import shutil
import tarfile
import tempfile
from pathlib import Path

from .error import SigMFFileError

SIGMF_ARCHIVE_EXT = ".sigmf"
SIGMF_METADATA_EXT = ".sigmf-meta"
SIGMF_DATASET_EXT = ".sigmf-data"
SIGMF_COLLECTION_EXT = ".sigmf-collection"


class SigMFArchive:
    """
    Archive a SigMFFile

    A `.sigmf` file must include both valid metadata and data.
    If `self.data_file` is not set or the requested output file
    is not writable, raises `SigMFFileError`.

    Parameters
    ----------

    sigmffile : SigMFFile
        A SigMFFile object with valid metadata and data_file.

    name : PathLike | str | bytes
        Path to archive file to create. If file exists, overwrite.
        If `name` doesn't end in .sigmf, it will be appended.
        For example: if `name` == "/tmp/archive1", then the
        following archive will be created:
            /tmp/archive1.sigmf
            - archive1/
                - archive1.sigmf-meta
                - archive1.sigmf-data

    fileobj : BufferedWriter
        If `fileobj` is specified, it is used as an alternative to
        a file object opened in binary mode for `name`. It is
        supposed to be at position 0. `name` is not required, but
        if specified will be used to determine the directory and
        file names within the archive. `fileobj` won't be closed.
        For example: if `name` == "archive1" and fileobj is given,
        a tar archive will be written to fileobj with the
        following structure:
            - archive1/
                - archive1.sigmf-meta
                - archive1.sigmf-data
    """

    def __init__(self, sigmffile, name=None, fileobj=None):
        is_buffer = fileobj is not None
        self.sigmffile = sigmffile
        self.path, arcname, fileobj = self._resolve(name, fileobj)

        self._ensure_data_file_set()
        self._validate()

        tar = tarfile.TarFile(mode="w", fileobj=fileobj, format=tarfile.PAX_FORMAT)
        tmpdir = Path(tempfile.mkdtemp())
        meta_path = tmpdir / (arcname + SIGMF_METADATA_EXT)
        data_path = tmpdir / (arcname + SIGMF_DATASET_EXT)

        # write files
        with open(meta_path, "w") as handle:
            self.sigmffile.dump(handle)
        if isinstance(self.sigmffile.data_buffer, io.BytesIO):
            # write data buffer to archive
            self.sigmffile.data_file = data_path
            with open(data_path, "wb") as handle:
                handle.write(self.sigmffile.data_buffer.getbuffer())
        else:
            # copy data to archive
            shutil.copy(self.sigmffile.data_file, data_path)
        tar.add(tmpdir, arcname=arcname, filter=self.chmod)
        # close files & remove tmpdir
        tar.close()
        if not is_buffer:
            # only close fileobj if we aren't working w/a buffer
            fileobj.close()
        shutil.rmtree(tmpdir)

    @staticmethod
    def chmod(tarinfo: tarfile.TarInfo):
        """permission filter for writing tar files"""
        if tarinfo.isdir():
            tarinfo.mode = 0o755  # dwrxw-rw-r
        else:
            tarinfo.mode = 0o644  # -wr-r--r--
        return tarinfo

    def _ensure_data_file_set(self):
        if not self.sigmffile.data_file and not isinstance(self.sigmffile.data_buffer, io.BytesIO):
            raise SigMFFileError("No data file in SigMFFile; use `set_data_file` before archiving.")

    def _validate(self):
        self.sigmffile.validate()

    def _resolve(self, name, fileobj):
        """
        Resolve both (name, fileobj) into (path, arcname, fileobj) given either or both.

        Returns
        -------
        path : PathLike
            Path of the archive file.
        arcname : str
            Name of the sigmf object within the archive.
        fileobj : BufferedWriter
            Open file handle object.
        """
        if fileobj:
            try:
                # exception if not byte-writable
                fileobj.write(bytes())
                # exception if no name property of handle
                path = Path(fileobj.name)
                if not name:
                    arcname = path.stem
                else:
                    arcname = name
            except io.UnsupportedOperation as exc:
                raise SigMFFileError(f"fileobj {fileobj} is not byte-writable.") from exc
            except AttributeError as exc:
                raise SigMFFileError(f"fileobj {fileobj} is invalid.") from exc
        elif name:
            path = Path(name)
            # ensure name has correct suffix if it exists
            if path.suffix == "":
                # add extension if none was given
                path = path.with_suffix(SIGMF_ARCHIVE_EXT)
            elif path.suffix != SIGMF_ARCHIVE_EXT:
                # ensure suffix is correct
                raise SigMFFileError(f"Invalid extension ({path.suffix} != {SIGMF_ARCHIVE_EXT}).")
            arcname = path.stem

            try:
                fileobj = open(path, "wb")
            except (OSError, IOError) as exc:
                raise SigMFFileError(f"Can't open {name} for writing.") from exc
        else:
            raise SigMFFileError("Either `name` or `fileobj` needs to be defined.")

        return path, arcname, fileobj



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/SigMF
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Access SigMF archives without extracting them."""

import io
import tarfile
from pathlib import Path

from . import __version__
from .archive import SIGMF_ARCHIVE_EXT, SIGMF_DATASET_EXT, SIGMF_METADATA_EXT
from .error import SigMFFileError
from .sigmffile import SigMFFile


class SigMFArchiveReader:
    """
    Access data within SigMF archive tarball in-place without extracting.

    Parameters
    ----------
    name : str | bytes | PathLike, optional
        Optional path to archive file to access.
    skip_checksum : bool, optional
        Skip dataset checksum calculation.
    map_readonly : bool, optional
        Indicate whether assignments on the numpy.memmap are allowed.
    archive_buffer : buffer, optional


    Raises
    ------
    SigMFError
        Archive file does not exist or is improperly formatted.
    ValueError
        If invalid arguments.
    ValidationError
        If metadata is invalid.
    """

    def __init__(self, name=None, skip_checksum=False, map_readonly=True, archive_buffer=None):
        if name is not None:
            path = Path(name)
            if path.suffix != SIGMF_ARCHIVE_EXT:
                err = "archive extension != {}".format(SIGMF_ARCHIVE_EXT)
                raise SigMFFileError(err)

            tar_obj = tarfile.open(path)

        elif archive_buffer is not None:
            tar_obj = tarfile.open(fileobj=archive_buffer, mode="r:")

        else:
            raise ValueError("Either `name` or `archive_buffer` must be not None.")

        json_contents = None
        data_offset = None
        data_size_bytes = None

        for memb in tar_obj.getmembers():
            if memb.isdir():  # memb.type == tarfile.DIRTYPE:
                # the directory structure will be reflected in the member name
                continue

            elif memb.isfile():  # memb.type == tarfile.REGTYPE:
                if memb.name.endswith(SIGMF_METADATA_EXT):
                    json_contents = memb.name
                    if data_offset is None:
                        # consider a warnings.warn() here; the datafile should be earlier in the
                        # archive than the metadata, so that updating it (like, adding an annotation)
                        # is fast.
                        pass
                    with tar_obj.extractfile(memb) as memb_fid:
                        json_contents = memb_fid.read()

                elif memb.name.endswith(SIGMF_DATASET_EXT):
                    data_offset = memb.offset_data
                    data_size_bytes = memb.size
                    with tar_obj.extractfile(memb) as memb_fid:
                        data_buffer = io.BytesIO(memb_fid.read())

                else:
                    print(f"A regular file {memb.name} was found but ignored in the archive")
            else:
                print(f"A member of type {memb.type} and name {memb.name} was found but not handled, just FYI.")

        if data_offset is None:
            raise SigMFFileError("No .sigmf-data file found in archive!")

        self.sigmffile = SigMFFile(metadata=json_contents)
        self.sigmffile.validate()

        self.sigmffile.set_data_file(
            data_buffer=data_buffer,
            skip_checksum=skip_checksum,
            size_bytes=data_size_bytes,
            map_readonly=map_readonly,
        )

        self.ndim = self.sigmffile.ndim
        self.shape = self.sigmffile.shape

        tar_obj.close()

    def __len__(self):
        return self.sigmffile.__len__()

    def __iter__(self):
        return self.sigmffile.__iter__()

    def __getitem__(self, sli):
        return self.sigmffile.__getitem__(sli)



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""converter for wav containers"""

import argparse
import getpass
import logging
import tempfile
from datetime import datetime, timezone
from os import PathLike
from pathlib import Path
from typing import Optional

from scipy.io import wavfile

from .. import SigMFFile
from .. import __version__ as toolversion
from ..sigmffile import get_sigmf_filenames
from ..utils import SIGMF_DATETIME_ISO8601_FMT, get_data_type_str

log = logging.getLogger()


def convert_wav(
    wav_path: str,
    out_path: Optional[str] = None,
    author: Optional[str] = None,
) -> PathLike:
    """
    Read a wav and write a sigmf archive.
    """
    wav_path = Path(wav_path)
    wav_stem = wav_path.stem
    samp_rate, wav_data = wavfile.read(wav_path)

    global_info = {
        SigMFFile.AUTHOR_KEY: getpass.getuser() if author is None else author,
        SigMFFile.DATATYPE_KEY: get_data_type_str(wav_data),
        SigMFFile.DESCRIPTION_KEY: f"converted from {wav_path.name}",
        SigMFFile.NUM_CHANNELS_KEY: 1 if len(wav_data.shape) < 2 else wav_data.shape[1],
        SigMFFile.RECORDER_KEY: "Official SigMF wav converter",
        SigMFFile.SAMPLE_RATE_KEY: samp_rate,
    }

    modify_time = wav_path.lstat().st_mtime
    wav_datetime = datetime.fromtimestamp(modify_time, tz=timezone.utc)

    capture_info = {
        SigMFFile.START_INDEX_KEY: 0,
        SigMFFile.DATETIME_KEY: wav_datetime.strftime(SIGMF_DATETIME_ISO8601_FMT),
    }

    temp_dir = Path(tempfile.mkdtemp())
    if out_path is None:
        # extension will be changed
        out_path = Path(wav_stem)
    else:
        out_path = Path(out_path)
    filenames = get_sigmf_filenames(out_path)

    data_path = temp_dir / filenames["data_fn"]
    wav_data.tofile(data_path)

    meta = SigMFFile(data_file=data_path, global_info=global_info)
    meta.add_capture(0, metadata=capture_info)
    log.debug("created %r", meta)

    arc_path = filenames["archive_fn"]
    meta.tofile(arc_path, toarchive=True)
    log.info("wrote %s", arc_path)
    return arc_path


def main() -> None:
    """
    entry-point for sigmf_convert_wav
    """
    parser = argparse.ArgumentParser(description="Convert wav to sigmf archive.")
    parser.add_argument("input", type=str, help="wav path")
    parser.add_argument("--author", type=str, default=None, help=f"set {SigMFFile.AUTHOR_KEY} metadata")
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument("--version", action="version", version=f"%(prog)s v{toolversion}")
    args = parser.parse_args()

    level_lut = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
    }
    logging.basicConfig(level=level_lut[min(args.verbose, 2)])

    _ = convert_wav(
        wav_path=args.input,
        author=args.author,
    )


if __name__ == "__main__":
    main()



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Provides pytest fixtures for other tests."""

import tempfile

import pytest

from sigmf import __specification__
from sigmf.archive import SIGMF_DATASET_EXT
from sigmf.sigmffile import SigMFFile

from .testdata import TEST_FLOAT32_DATA, TEST_METADATA


@pytest.fixture
def test_data_file():
    """when called, yields temporary dataset"""
    with tempfile.NamedTemporaryFile(suffix=f".{SIGMF_DATASET_EXT}") as temp:
        TEST_FLOAT32_DATA.tofile(temp.name)
        yield temp


@pytest.fixture
def test_sigmffile(test_data_file):
    """If pytest uses this signature, will return valid SigMF file."""
    meta = SigMFFile()
    meta.set_global_field("core:datatype", "rf32_le")
    meta.set_global_field("core:version", __specification__)
    meta.add_annotation(start_index=0, length=len(TEST_FLOAT32_DATA))
    meta.add_capture(start_index=0)
    meta.set_data_file(test_data_file.name)
    assert meta._metadata == TEST_METADATA
    return meta



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Tests for SigMFArchive"""

import codecs
import json
import tarfile
import tempfile
from pathlib import Path

import jsonschema
import numpy as np
import pytest

from sigmf import error
from sigmf.archive import SIGMF_DATASET_EXT, SIGMF_METADATA_EXT

from .testdata import TEST_FLOAT32_DATA, TEST_METADATA


def create_test_archive(test_sigmffile, tmpfile):
    sigmf_archive = test_sigmffile.archive(fileobj=tmpfile)
    sigmf_tarfile = tarfile.open(sigmf_archive, mode="r", format=tarfile.PAX_FORMAT)
    return sigmf_tarfile


def test_without_data_file_throws_fileerror(test_sigmffile):
    test_sigmffile.data_file = None
    with tempfile.NamedTemporaryFile() as temp:
        with pytest.raises(error.SigMFFileError):
            test_sigmffile.archive(name=temp.name)


def test_invalid_md_throws_validationerror(test_sigmffile):
    del test_sigmffile._metadata["global"]["core:datatype"]  # required field
    with tempfile.NamedTemporaryFile() as temp:
        with pytest.raises(jsonschema.exceptions.ValidationError):
            test_sigmffile.archive(name=temp.name)


def test_name_wrong_extension_throws_fileerror(test_sigmffile):
    with tempfile.NamedTemporaryFile() as temp:
        with pytest.raises(error.SigMFFileError):
            test_sigmffile.archive(name=temp.name + ".zip")


def test_fileobj_extension_ignored(test_sigmffile):
    with tempfile.NamedTemporaryFile(suffix=".tar") as temp:
        test_sigmffile.archive(fileobj=temp)


def test_name_used_in_fileobj(test_sigmffile):
    with tempfile.NamedTemporaryFile() as temp:
        sigmf_archive = test_sigmffile.archive(name="testarchive", fileobj=temp)
        sigmf_tarfile = tarfile.open(sigmf_archive, mode="r")
        basedir, file1, file2 = sigmf_tarfile.getmembers()
        assert basedir.name == "testarchive"

        def filename(tarinfo):
            return Path(tarinfo.name).stem

        assert filename(file1) == "testarchive"
        assert filename(file2) == "testarchive"


def test_fileobj_not_closed(test_sigmffile):
    with tempfile.NamedTemporaryFile() as temp:
        test_sigmffile.archive(fileobj=temp)
        assert not temp.file.closed


def test_unwritable_fileobj_throws_fileerror(test_sigmffile):
    with tempfile.NamedTemporaryFile(mode="rb") as temp:
        with pytest.raises(error.SigMFFileError):
            test_sigmffile.archive(fileobj=temp)


def test_unwritable_name_throws_fileerror(test_sigmffile):
    # Cannot assume /root/ is unwritable (e.g. Docker environment)
    # so use invalid filename
    unwritable_file = "/bad_name/"
    with pytest.raises(error.SigMFFileError):
        test_sigmffile.archive(name=unwritable_file)


def test_tarfile_layout(test_sigmffile):
    with tempfile.NamedTemporaryFile() as temp:
        sigmf_tarfile = create_test_archive(test_sigmffile, temp)
        basedir, file1, file2 = sigmf_tarfile.getmembers()
        assert tarfile.TarInfo.isdir(basedir)
        assert tarfile.TarInfo.isfile(file1)
        assert tarfile.TarInfo.isfile(file2)


def test_tarfile_names_and_extensions(test_sigmffile):
    with tempfile.NamedTemporaryFile() as temp:
        sigmf_tarfile = create_test_archive(test_sigmffile, temp)
        basedir, file1, file2 = sigmf_tarfile.getmembers()
        archive_name = basedir.name
        assert archive_name == Path(temp.name).name
        file_extensions = {SIGMF_DATASET_EXT, SIGMF_METADATA_EXT}

        file1_name, file1_ext = Path(file1.name).stem, Path(file1.name).suffix
        assert file1_name == archive_name
        assert file1_ext in file_extensions

        file_extensions.remove(file1_ext)

        file2_name, file2_ext = Path(file2.name).stem, Path(file2.name).suffix
        assert file2_name == archive_name
        assert file2_ext in file_extensions


def test_tarfile_persmissions(test_sigmffile):
    with tempfile.NamedTemporaryFile() as temp:
        sigmf_tarfile = create_test_archive(test_sigmffile, temp)
        basedir, file1, file2 = sigmf_tarfile.getmembers()
        assert basedir.mode == 0o755
        assert file1.mode == 0o644
        assert file2.mode == 0o644


def test_contents(test_sigmffile):
    with tempfile.NamedTemporaryFile() as temp:
        sigmf_tarfile = create_test_archive(test_sigmffile, temp)
        basedir, file1, file2 = sigmf_tarfile.getmembers()
        if file1.name.endswith(SIGMF_METADATA_EXT):
            mdfile = file1
            datfile = file2
        else:
            mdfile = file2
            datfile = file1

        bytestream_reader = codecs.getreader("utf-8")  # bytes -> str
        mdfile_reader = bytestream_reader(sigmf_tarfile.extractfile(mdfile))
        assert json.load(mdfile_reader) == TEST_METADATA

        datfile_reader = sigmf_tarfile.extractfile(datfile)
        # calling `fileno` on `tarfile.ExFileObject` throws error (?), but
        # np.fromfile requires it, so we need this extra step
        data = np.frombuffer(datfile_reader.read(), dtype=np.float32)

        assert np.array_equal(data, TEST_FLOAT32_DATA)


def test_tarfile_type(test_sigmffile):
    with tempfile.NamedTemporaryFile() as temp:
        sigmf_tarfile = create_test_archive(test_sigmffile, temp)
        assert sigmf_tarfile.format == tarfile.PAX_FORMAT



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Hashing Functions"""

import hashlib
from pathlib import Path


def calculate_sha512(filename=None, fileobj=None, offset=None, size=None):
    """
    Return sha512 of file or fileobj.
    """
    the_hash = hashlib.sha512()
    bytes_to_hash = size
    bytes_read = 0

    if filename is not None:
        fileobj = open(filename, "rb")
    if size is None:
        bytes_to_hash = Path(filename).stat().st_size
    else:
        fileobj.seek(offset)

    while bytes_read < bytes_to_hash:
        buff = fileobj.read(min(4096, (bytes_to_hash - bytes_read)))
        the_hash.update(buff)
        bytes_read += len(buff)

    if filename is not None:
        fileobj.close()

    return the_hash.hexdigest()



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""SigMF Validator"""
import argparse
import glob
import json
import logging
import os
import sys

# multi-threading library - should work well as I/O will be the primary
# cost for small SigMF files. Swap to ProcessPool if files are large.
from concurrent.futures import ThreadPoolExecutor, as_completed

# required for Python 3.7
from typing import Optional, Tuple

import jsonschema

from . import __version__ as toolversion
from . import error, schema, sigmffile


def validate(metadata, ref_schema=schema.get_schema()) -> None:
    """
    Check that the provided `metadata` dict is valid according to the `ref_schema` dict.
    Walk entire schema and check all keys.

    Parameters
    ----------
    metadata : dict
        The SigMF metadata to be validated.
    ref_schema : dict, optional
        The schema that holds the SigMF metadata definition.
        Since the schema evolves over time, we may want to be able to check
        against different versions in the *future*.

    Raises
    ------
    ValidationError
        If metadata is invalid.
    """
    jsonschema.validators.validate(instance=metadata, schema=ref_schema)

    # ensure captures and annotations have monotonically increasing sample_start
    for key in ["captures", "annotations"]:
        count = -1
        for item in metadata[key]:
            new_count = item["core:sample_start"]
            if new_count < count:
                raise jsonschema.exceptions.ValidationError(f"{key} has incorrect sample start ordering.")
            count = new_count


def _validate_single_file(filename, skip_checksum: bool, logger: logging.Logger) -> int:
    """Validates a single SigMF file.

    To be called as part of a multithreading / multiprocess application.

    Parameters
    ----------
    filename : str
        Path and name to sigmf.data or sigmf.meta file.
    skip_checksum : bool
        Whether to perform checksum computation.
    logger : logging.Logger
        Logging object to log errors to.

    Returns
    -------
    rc : int
        0 if OK, 1 if err
    """
    try:
        # load signal
        signal = sigmffile.fromfile(filename, skip_checksum=skip_checksum)
        # validate
        signal.validate()

    # handle any of 4 exceptions at once...
    except (jsonschema.exceptions.ValidationError, error.SigMFFileError, json.decoder.JSONDecodeError, IOError) as err:
        # catch the error, log, and continue
        logger.error(f"file `{filename}`: {err}")
        return 1
    else:
        return 0


def main(arg_tuple: Optional[Tuple[str, ...]] = None) -> None:
    """entry-point for command-line validator"""
    parser = argparse.ArgumentParser(
        description="Validate SigMF Archive or file pair against JSON schema.", prog="sigmf_validate"
    )
    parser.add_argument("path", nargs="*", help="SigMF path(s). Accepts * wildcards and extensions are optional.")
    parser.add_argument("--skip-checksum", action="store_true", help="Skip reading dataset to validate checksum.")
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument("--version", action="version", version=f"%(prog)s {toolversion}")

    # allow pass-in arg_tuple for testing purposes
    args = parser.parse_args(arg_tuple)

    level_lut = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
    }
    log = logging.getLogger()
    logging.basicConfig(level=level_lut[min(args.verbose, 2)])

    paths = []
    # resolve possible wildcards
    for path in args.path:
        paths += glob.glob(path)

    # multi-processing / threading pathway.
    n_completed = 0
    n_total = len(paths)
    # estimate number of CPU cores
    # https://stackoverflow.com/questions/1006289/how-to-find-out-the-number-of-cpus-using-python
    est_num_workers = len(os.sched_getaffinity(0)) if os.name == 'posix' else os.cpu_count()
    # create a thread pool
    # https://docs.python.org/3.7/library/concurrent.futures.html#threadpoolexecutor
    with ThreadPoolExecutor(max_workers=est_num_workers) as executor:
        # submit jobs
        future_validations = {executor.submit(_validate_single_file, path, args.skip_checksum, log) for path in paths}
        # load and await jobs to complete... no return
        for future in as_completed(future_validations):
            if future.result() == 0:
                n_completed += 1

    if n_total == 0:
        log.error("No paths to validate.")
        sys.exit(1)
    elif n_completed != n_total:
        log.info("Validated %d of %d files OK", n_completed, n_total)
        sys.exit(1)
    else:
        log.info("Validated all %d files OK!", n_total)


if __name__ == "__main__":
    main()



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Tests for Non-Conforming Datasets"""

import copy
import shutil
import tempfile
import unittest
from pathlib import Path

import numpy as np
from hypothesis import given
from hypothesis import strategies as st

from sigmf.error import SigMFFileError
from sigmf.sigmffile import SigMFFile, fromfile

from .testdata import TEST_FLOAT32_DATA, TEST_METADATA


class TestNonConformingDataset(unittest.TestCase):
    """unit tests for NCD"""

    def setUp(self):
        """create temporary path"""
        self.temp_dir = Path(tempfile.mkdtemp())

    def tearDown(self):
        """remove temporary path"""
        shutil.rmtree(self.temp_dir)

    @given(st.sampled_from([".", "subdir/", "sub0/sub1/sub2/"]))
    def test_load_ncd(self, subdir: str) -> None:
        """test loading non-conforming dataset"""
        data_path = self.temp_dir / subdir / "dat.bin"
        meta_path = self.temp_dir / subdir / "dat.sigmf-meta"
        Path.mkdir(data_path.parent, parents=True, exist_ok=True)

        # create data file
        TEST_FLOAT32_DATA.tofile(data_path)

        # create metadata file
        ncd_metadata = copy.deepcopy(TEST_METADATA)
        meta = SigMFFile(metadata=ncd_metadata, data_file=data_path)
        meta.tofile(meta_path)

        # load dataset & validate we can read all the data
        meta_loopback = fromfile(meta_path)
        self.assertTrue(np.array_equal(TEST_FLOAT32_DATA, meta_loopback.read_samples()))
        self.assertTrue(np.array_equal(TEST_FLOAT32_DATA, meta_loopback[:]))

        # delete the non-conforming dataset and ensure error is raised due to missing dataset;
        # in Windows the SigMFFile instances need to be garbage collected first,
        # otherwise the np.memmap instances (stored in self._memmap) block the deletion
        meta = None
        meta_loopback = None
        Path.unlink(data_path)
        with self.assertRaises(SigMFFileError):
            _ = fromfile(meta_path)



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Tests for SigMFFile Object"""

import copy
import json
import shutil
import tempfile
import unittest
from pathlib import Path

import numpy as np

from sigmf import error, sigmffile, utils
from sigmf.sigmffile import SigMFFile

from .testdata import *


class TestClassMethods(unittest.TestCase):
    def setUp(self):
        """ensure tests have a valid SigMF object to work with"""
        self.temp_dir = Path(tempfile.mkdtemp())
        self.temp_path_data = self.temp_dir / "trash.sigmf-data"
        self.temp_path_meta = self.temp_dir / "trash.sigmf-meta"
        TEST_FLOAT32_DATA.tofile(self.temp_path_data)
        self.sigmf_object = SigMFFile(TEST_METADATA, data_file=self.temp_path_data)
        self.sigmf_object.tofile(self.temp_path_meta)

    def tearDown(self):
        """remove temporary dir"""
        shutil.rmtree(self.temp_dir)

    def test_pathlib_handle(self):
        """ensure file can be a string or a pathlib object"""
        self.assertTrue(self.temp_path_data.exists())
        obj_str = sigmffile.fromfile(str(self.temp_path_data))
        obj_str.validate()
        obj_pth = sigmffile.fromfile(self.temp_path_data)
        obj_pth.validate()

    def test_filenames_with_dots(self):
        """test that filenames with non-extension . characters are handled correctly"""
        filenames = ["a", "b.c", "d.e.f"]
        for filename in filenames:
            temp_path_data = self.temp_dir / f"{filename}.sigmf-data"
            temp_path_meta = self.temp_dir / f"{filename}.sigmf-meta"
            TEST_FLOAT32_DATA.tofile(temp_path_data)
            self.sigmf_object = SigMFFile(TEST_METADATA, data_file=temp_path_data)
            self.sigmf_object.tofile(temp_path_meta)
            files = [str(temp_path_data), temp_path_data, str(temp_path_meta), temp_path_meta]
            for filename in files:
                obj = sigmffile.fromfile(filename)
                obj.validate()

    def test_iterator_basic(self):
        """make sure default batch_size works"""
        count = 0
        for _ in self.sigmf_object:
            count += 1
        self.assertEqual(count, len(self.sigmf_object))

    def test_checksum(self):
        """Ensure checksum fails when incorrect or empty string."""
        for new_checksum in ("", "a", 0):
            bad_checksum_metadata = copy.deepcopy(TEST_METADATA)
            bad_checksum_metadata[SigMFFile.GLOBAL_KEY][SigMFFile.HASH_KEY] = new_checksum
            with self.assertRaises(error.SigMFFileError):
                _ = SigMFFile(bad_checksum_metadata, self.temp_path_data)

    def test_equality(self):
        """Ensure __eq__ working as expected"""
        other = SigMFFile(copy.deepcopy(TEST_METADATA))
        self.assertEqual(self.sigmf_object, other)
        # different after changing any part of metadata
        other.add_annotation(start_index=0, metadata={"a": 0})
        self.assertNotEqual(self.sigmf_object, other)


class TestAnnotationHandling(unittest.TestCase):
    def test_get_annotations_with_index(self):
        """Test that only annotations containing index are returned from get_annotations()"""
        smf = SigMFFile(copy.deepcopy(TEST_METADATA))
        smf.add_annotation(start_index=1)
        smf.add_annotation(start_index=4, length=4)
        annotations_idx10 = smf.get_annotations(index=10)
        self.assertListEqual(
            annotations_idx10,
            [
                {SigMFFile.START_INDEX_KEY: 0, SigMFFile.LENGTH_INDEX_KEY: 16},
                {SigMFFile.START_INDEX_KEY: 1},
            ],
        )

    def test__count_samples_from_annotation(self):
        """Make sure sample count from annotations use correct end index"""
        smf = SigMFFile(copy.deepcopy(TEST_METADATA))
        smf.add_annotation(start_index=0, length=32)
        smf.add_annotation(start_index=4, length=4)
        sample_count = smf._count_samples()
        self.assertEqual(sample_count, 32)

    def test_set_data_file_without_annotations(self):
        """
        Make sure setting data_file with no annotations registered does not
        raise any errors
        """
        smf = SigMFFile(copy.deepcopy(TEST_METADATA))
        smf._metadata[SigMFFile.ANNOTATION_KEY].clear()
        with tempfile.TemporaryDirectory() as tmpdir:
            temp_path_data = Path(tmpdir) / "datafile"
            TEST_FLOAT32_DATA.tofile(temp_path_data)
            smf.set_data_file(temp_path_data)
            samples = smf.read_samples()
            self.assertTrue(len(samples) == 16)

    def test_set_data_file_with_annotations(self):
        """
        Make sure setting data_file with annotations registered use sample
        count from data_file and issue a warning if annotations have end
        indices bigger than file end index
        """
        smf = SigMFFile(copy.deepcopy(TEST_METADATA))
        smf.add_annotation(start_index=0, length=32)
        with tempfile.TemporaryDirectory() as tmpdir:
            temp_path_data = Path(tmpdir) / "datafile"
            TEST_FLOAT32_DATA.tofile(temp_path_data)
            with self.assertWarns(Warning):
                # Issues warning since file ends before the final annotatio
                smf.set_data_file(temp_path_data)
                samples = smf.read_samples()
                self.assertTrue(len(samples) == 16)


class TestMultichannel(unittest.TestCase):
    def setUp(self):
        # in order to check shapes we need some positive number of samples to work with
        # number of samples should be lowest common factor of num_channels
        self.raw_count = 16
        self.lut = {
            "i8": np.int8,
            "u8": np.uint8,
            "i16": np.int16,
            "u16": np.uint16,
            "u32": np.uint32,
            "i32": np.int32,
            "f32": np.float32,
            "f64": np.float64,
        }
        self.temp_file = tempfile.NamedTemporaryFile()
        self.temp_path = Path(self.temp_file.name)

    def tearDown(self):
        """clean-up temporary files"""
        self.temp_file.close()

    def test_multichannel_types(self):
        """check that real & complex for all types is reading multiple channels correctly"""
        for key, dtype in self.lut.items():
            # for each type of storage
            np.arange(self.raw_count, dtype=dtype).tofile(self.temp_path)
            for num_channels in [1, 4, 8]:
                # for single or 8 channel
                for complex_prefix in ["r", "c"]:
                    # for real or complex
                    check_count = self.raw_count
                    temp_signal = SigMFFile(
                        data_file=self.temp_path,
                        global_info={
                            SigMFFile.DATATYPE_KEY: f"{complex_prefix}{key}_le",
                            SigMFFile.NUM_CHANNELS_KEY: num_channels,
                        },
                    )
                    temp_samples = temp_signal.read_samples()

                    if complex_prefix == "c":
                        # complex data will be half as long
                        check_count //= 2
                        self.assertTrue(np.all(np.iscomplex(temp_samples)))
                    if num_channels != 1:
                        self.assertEqual(temp_samples.ndim, 2)
                    check_count //= num_channels

                    self.assertEqual(check_count, temp_signal._count_samples())

    def test_multichannel_seek(self):
        """ensure that seeking is working correctly with multichannel files"""
        # write some dummy data and read back
        np.arange(18, dtype=np.uint16).tofile(self.temp_path)
        temp_signal = SigMFFile(
            data_file=self.temp_path,
            global_info={
                SigMFFile.DATATYPE_KEY: "cu16_le",
                SigMFFile.NUM_CHANNELS_KEY: 3,
            },
        )
        # read after the first sample
        temp_samples = temp_signal.read_samples(start_index=1, autoscale=False)
        # ensure samples are in the order we expect
        self.assertTrue(np.all(temp_samples[:, 0] == np.array([6 + 7j, 12 + 13j])))


def test_key_validity():
    """ensure the keys in test metadata are valid"""
    for top_key, top_val in TEST_METADATA.items():
        if isinstance(top_val, dict):
            for core_key in top_val.keys():
                assert core_key in vars(SigMFFile)[f"VALID_{top_key.upper()}_KEYS"]
        elif isinstance(top_val, list):
            # annotations are in a list
            for annot in top_val:
                for core_key in annot.keys():
                    assert core_key in SigMFFile.VALID_ANNOTATION_KEYS
        else:
            raise ValueError("expected list or dict")


def test_ordered_metadata():
    """check to make sure the metadata is sorted as expected"""
    sigf = SigMFFile()
    top_sort_order = ["global", "captures", "annotations"]
    for kdx, key in enumerate(sigf.ordered_metadata()):
        assert kdx == top_sort_order.index(key)


class TestCaptures(unittest.TestCase):
    """ensure capture access tools work properly"""

    def setUp(self) -> None:
        """ensure tests have a valid SigMF object to work with"""
        self.temp_dir = Path(tempfile.mkdtemp())
        self.temp_path_data = self.temp_dir / "trash.sigmf-data"
        self.temp_path_meta = self.temp_dir / "trash.sigmf-meta"

    def tearDown(self) -> None:
        """remove temporary dir"""
        shutil.rmtree(self.temp_dir)

    def prepare(self, data: list, meta: dict, dtype: type) -> SigMFFile:
        """write some data and metadata to temporary paths"""
        np.array(data, dtype=dtype).tofile(self.temp_path_data)
        with open(self.temp_path_meta, "w") as handle:
            json.dump(meta, handle)
        meta = sigmffile.fromfile(self.temp_path_meta, skip_checksum=True)
        return meta

    def test_000(self) -> None:
        """compliant two-capture recording"""
        meta = self.prepare(TEST_U8_DATA0, TEST_U8_META0, np.uint8)
        self.assertEqual(256, meta._count_samples())
        self.assertTrue(meta._is_conforming_dataset())
        self.assertTrue((0, 0), meta.get_capture_byte_boundarys(0))
        self.assertTrue((0, 256), meta.get_capture_byte_boundarys(1))
        self.assertTrue(np.array_equal(TEST_U8_DATA0, meta.read_samples(autoscale=False)))
        self.assertTrue(np.array_equal(np.array([]), meta.read_samples_in_capture(0)))
        self.assertTrue(np.array_equal(TEST_U8_DATA0, meta.read_samples_in_capture(1, autoscale=False)))

    def test_001(self) -> None:
        """two capture recording with header_bytes and trailing_bytes set"""
        meta = self.prepare(TEST_U8_DATA1, TEST_U8_META1, np.uint8)
        self.assertEqual(192, meta._count_samples())
        self.assertFalse(meta._is_conforming_dataset())
        self.assertTrue((32, 160), meta.get_capture_byte_boundarys(0))
        self.assertTrue((160, 224), meta.get_capture_byte_boundarys(1))
        self.assertTrue(np.array_equal(np.arange(128), meta.read_samples_in_capture(0, autoscale=False)))
        self.assertTrue(np.array_equal(np.arange(128, 192), meta.read_samples_in_capture(1, autoscale=False)))

    def test_002(self) -> None:
        """two capture recording with multiple header_bytes set"""
        meta = self.prepare(TEST_U8_DATA2, TEST_U8_META2, np.uint8)
        self.assertEqual(192, meta._count_samples())
        self.assertFalse(meta._is_conforming_dataset())
        self.assertTrue((32, 160), meta.get_capture_byte_boundarys(0))
        self.assertTrue((176, 240), meta.get_capture_byte_boundarys(1))
        self.assertTrue(np.array_equal(np.arange(128), meta.read_samples_in_capture(0, autoscale=False)))
        self.assertTrue(np.array_equal(np.arange(128, 192), meta.read_samples_in_capture(1, autoscale=False)))

    def test_003(self) -> None:
        """three capture recording with multiple header_bytes set"""
        meta = self.prepare(TEST_U8_DATA3, TEST_U8_META3, np.uint8)
        self.assertEqual(192, meta._count_samples())
        self.assertFalse(meta._is_conforming_dataset())
        self.assertTrue((32, 64), meta.get_capture_byte_boundarys(0))
        self.assertTrue((64, 160), meta.get_capture_byte_boundarys(1))
        self.assertTrue((192, 256), meta.get_capture_byte_boundarys(2))
        self.assertTrue(np.array_equal(np.arange(32), meta.read_samples_in_capture(0, autoscale=False)))
        self.assertTrue(np.array_equal(np.arange(32, 128), meta.read_samples_in_capture(1, autoscale=False)))
        self.assertTrue(np.array_equal(np.arange(128, 192), meta.read_samples_in_capture(2, autoscale=False)))

    def test_004(self) -> None:
        """two channel version of 000"""
        meta = self.prepare(TEST_U8_DATA4, TEST_U8_META4, np.uint8)
        self.assertEqual(96, meta._count_samples())
        self.assertFalse(meta._is_conforming_dataset())
        self.assertTrue((32, 160), meta.get_capture_byte_boundarys(0))
        self.assertTrue((160, 224), meta.get_capture_byte_boundarys(1))
        self.assertTrue(
            np.array_equal(np.arange(64).repeat(2).reshape(-1, 2), meta.read_samples_in_capture(0, autoscale=False))
        )
        self.assertTrue(
            np.array_equal(np.arange(64, 96).repeat(2).reshape(-1, 2), meta.read_samples_in_capture(1, autoscale=False))
        )

    def test_slicing_ru8(self) -> None:
        """slice real uint8"""
        meta = self.prepare(TEST_U8_DATA0, TEST_U8_META0, np.uint8)
        self.assertTrue(np.array_equal(meta[:], TEST_U8_DATA0))
        self.assertTrue(np.array_equal(meta[6], TEST_U8_DATA0[6]))
        self.assertTrue(np.array_equal(meta[1:-1], TEST_U8_DATA0[1:-1]))

    def test_slicing_rf32(self) -> None:
        """slice real float32"""
        meta = self.prepare(TEST_FLOAT32_DATA, TEST_METADATA, np.float32)
        self.assertTrue(np.array_equal(meta[:], TEST_FLOAT32_DATA))
        self.assertTrue(np.array_equal(meta[9], TEST_FLOAT32_DATA[9]))

    def test_slicing_multiple_channels(self) -> None:
        """slice multiple channels"""
        meta = self.prepare(TEST_U8_DATA4, TEST_U8_META4, np.uint8)
        channelized = np.array(TEST_U8_DATA4).reshape((-1, 2))
        self.assertTrue(np.array_equal(meta[:][:], channelized))
        self.assertTrue(np.array_equal(meta[10:20, 0], meta.read_samples(autoscale=False)[10:20, 0]))
        self.assertTrue(np.array_equal(meta[0], channelized[0]))
        self.assertTrue(np.array_equal(meta[1, :], channelized[1]))


def simulate_capture(sigmf_md, n, capture_len):
    start_index = capture_len * n

    capture_md = {"core:datetime": utils.get_sigmf_iso8601_datetime_now()}

    sigmf_md.add_capture(start_index=start_index, metadata=capture_md)

    annotation_md = {
        "core:latitude": 40.0 + 0.0001 * n,
        "core:longitude": -105.0 + 0.0001 * n,
    }

    sigmf_md.add_annotation(start_index=start_index, length=capture_len, metadata=annotation_md)


def test_default_constructor():
    SigMFFile()


def test_set_non_required_global_field():
    sigf = SigMFFile()
    sigf.set_global_field("this_is:not_in_the_schema", None)


def test_add_capture():
    sigf = SigMFFile()
    sigf.add_capture(start_index=0, metadata={})


def test_add_annotation():
    sigf = SigMFFile()
    sigf.add_capture(start_index=0)
    meta = {"latitude": 40.0, "longitude": -105.0}
    sigf.add_annotation(start_index=0, length=128, metadata=meta)


def test_fromarchive(test_sigmffile):
    with tempfile.NamedTemporaryFile(suffix=".sigmf") as temp_file:
        archive_path = test_sigmffile.archive(name=temp_file.name)
        result = sigmffile.fromarchive(archive_path=archive_path)
        assert result._metadata == test_sigmffile._metadata == TEST_METADATA


def test_add_multiple_captures_and_annotations():
    sigf = SigMFFile()
    for idx in range(3):
        simulate_capture(sigf, idx, 1024)



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Tests for Utilities"""

from datetime import datetime, timezone

import pytest

from sigmf import utils


# fmt: off
@pytest.mark.parametrize("time_str, expected", [
    ("1955-07-04T05:15:00Z",                datetime(year=1955, month=7,  day=4, hour=5,  minute=15, second=00, microsecond=0, tzinfo=timezone.utc)),
    ("2956-08-05T06:15:12Z",                datetime(year=2956, month=8,  day=5, hour=6,  minute=15, second=12, microsecond=0, tzinfo=timezone.utc)),
    ("3957-09-06T07:15:12.345Z",            datetime(year=3957, month=9,  day=6, hour=7,  minute=15, second=12, microsecond=345000, tzinfo=timezone.utc)),
    ("4958-10-07T08:15:12.0345Z",           datetime(year=4958, month=10, day=7, hour=8,  minute=15, second=12, microsecond=34500, tzinfo=timezone.utc)),
    ("5959-11-08T09:15:12.000000Z",         datetime(year=5959, month=11, day=8, hour=9,  minute=15, second=12, microsecond=0, tzinfo=timezone.utc)),
    ("6960-12-09T10:15:12.123456789123Z",   datetime(year=6960, month=12, day=9, hour=10, minute=15, second=12, microsecond=123456, tzinfo=timezone.utc)),
])
# fmt: on
def test_parse_simple_iso8601(time_str: str, expected: datetime) -> None:
    """Ensure various times are represented as expected"""
    date_struct = utils.parse_iso8601_datetime(time_str)
    assert date_struct == expected


def test_roundtrip_datetime() -> None:
    """New string -> struct -> string is ok"""
    now_str = utils.get_sigmf_iso8601_datetime_now()
    now_struct = utils.parse_iso8601_datetime(now_str)
    assert now_str == now_struct.strftime(utils.SIGMF_DATETIME_ISO8601_FMT)



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Schema IO"""

import json
from pathlib import Path

from . import __version__ as toolversion

SCHEMA_META = "schema-meta.json"
SCHEMA_COLLECTION = "schema-collection.json"


def get_schema(version=toolversion, schema_file=SCHEMA_META):
    """
    Load JSON Schema to for either a `sigmf-meta` or `sigmf-collection`.

    TODO: In the future load specific schema versions.
    """
    schema_dir = Path(__file__).parent
    with open(schema_dir / schema_file, "rb") as handle:
        schema = json.load(handle)
    return schema



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Tests for collections"""

import copy
import os
import shutil
import tempfile
import unittest
from pathlib import Path

import numpy as np
from hypothesis import given
from hypothesis import strategies as st

from sigmf.archive import SIGMF_COLLECTION_EXT, SIGMF_DATASET_EXT, SIGMF_METADATA_EXT
from sigmf.sigmffile import SigMFCollection, SigMFFile, fromfile

from .testdata import TEST_FLOAT32_DATA, TEST_METADATA


class TestCollection(unittest.TestCase):
    """unit tests for colections"""

    def setUp(self):
        """create temporary path"""
        self.temp_dir = Path(tempfile.mkdtemp())

    def tearDown(self):
        """remove temporary path"""
        shutil.rmtree(self.temp_dir)

    @given(st.sampled_from([".", "subdir/", "sub0/sub1/sub2/"]))
    def test_load_collection(self, subdir: str) -> None:
        """test path handling for collections"""
        data_name1 = "dat1" + SIGMF_DATASET_EXT
        data_name2 = "dat2" + SIGMF_DATASET_EXT
        meta_name1 = "dat1" + SIGMF_METADATA_EXT
        meta_name2 = "dat2" + SIGMF_METADATA_EXT
        collection_name = "collection" + SIGMF_COLLECTION_EXT
        data_path1 = self.temp_dir / subdir / data_name1
        data_path2 = self.temp_dir / subdir / data_name2
        meta_path1 = self.temp_dir / subdir / meta_name1
        meta_path2 = self.temp_dir / subdir / meta_name2
        collection_path = self.temp_dir / subdir / collection_name
        os.makedirs(collection_path.parent, exist_ok=True)

        # create data files
        TEST_FLOAT32_DATA.tofile(data_path1)
        TEST_FLOAT32_DATA.tofile(data_path2)

        # create metadata files
        metadata = copy.deepcopy(TEST_METADATA)
        meta1 = SigMFFile(metadata=metadata, data_file=data_path1)
        meta2 = SigMFFile(metadata=metadata, data_file=data_path2)
        meta1.tofile(meta_path1)
        meta2.tofile(meta_path2)

        # create collection
        collection = SigMFCollection(
            metafiles=[meta_name1, meta_name2],
            base_path=str(self.temp_dir / subdir),
        )
        collection.tofile(collection_path)

        # load collection
        collection_loopback = fromfile(collection_path)
        meta1_loopback = collection_loopback.get_SigMFFile(stream_index=0)
        meta2_loopback = collection_loopback.get_SigMFFile(stream_index=1)

        self.assertTrue(np.array_equal(TEST_FLOAT32_DATA, meta1_loopback.read_samples()))
        self.assertTrue(np.array_equal(TEST_FLOAT32_DATA, meta2_loopback[:]))



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Tests for Validator"""

import tempfile
import unittest
from pathlib import Path

from jsonschema.exceptions import ValidationError

import sigmf
from sigmf import SigMFFile

from .testdata import TEST_FLOAT32_DATA, TEST_METADATA


def test_valid_data():
    """ensure the default metadata is OK"""
    SigMFFile(TEST_METADATA).validate()


class CommandLineValidator(unittest.TestCase):
    """Check behavior of command-line parser"""

    def setUp(self):
        """Create a directory with some valid files"""
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.tmp_path = tmp_path = Path(self.tmp_dir.name)
        junk_path = tmp_path / "junk"
        TEST_FLOAT32_DATA.tofile(junk_path)
        some_meta = SigMFFile(TEST_METADATA, data_file=junk_path)
        some_meta.tofile(tmp_path / "a")
        some_meta.tofile(tmp_path / "b")
        some_meta.tofile(tmp_path / "c", toarchive=True)

    def tearDown(self):
        """cleanup"""
        self.tmp_dir.cleanup()

    def test_normal(self):
        """able to parse archives and non-archives"""
        args = (str(self.tmp_path / "*.sigmf*"),)
        sigmf.validate.main(args)

    def test_normal_skip(self):
        """able to skip checksum"""
        args = (str(self.tmp_path / "*.sigmf*"), "--skip-checksum")
        sigmf.validate.main(args)

    def test_partial(self):
        """checks some but not all files"""
        args = (str(self.tmp_path / "*"),)
        with self.assertRaises(SystemExit):
            sigmf.validate.main(args)

    def test_missing(self):
        """exit with rc=1 when run on empty"""
        with self.assertRaises(SystemExit) as cm:
            sigmf.validate.main(tuple())
        self.assertEqual((1,), cm.exception.args)

    def test_version(self):
        """exit with rc=0 after printing version"""
        args = ("--version",)
        with self.assertRaises(SystemExit) as cm:
            sigmf.validate.main(args)
        self.assertEqual((0,), cm.exception.args)


class FailingCases(unittest.TestCase):
    """Cases where the validator should raise an exception."""

    def setUp(self):
        self.metadata = dict(TEST_METADATA)

    def test_no_version(self):
        """core:version must be present"""
        del self.metadata[SigMFFile.GLOBAL_KEY][SigMFFile.VERSION_KEY]
        with self.assertRaises(ValidationError):
            SigMFFile(self.metadata).validate()

    def test_extra_top_level_key(self):
        """no extra keys allowed on the top level"""
        self.metadata["extra"] = 0
        with self.assertRaises(ValidationError):
            SigMFFile(self.metadata).validate()

    def test_invalid_type(self):
        """license key must be string"""
        self.metadata[SigMFFile.GLOBAL_KEY][SigMFFile.LICENSE_KEY] = 1
        with self.assertRaises(ValidationError):
            SigMFFile(self.metadata).validate()

    def test_invalid_capture_order(self):
        """metadata must have captures in order"""
        self.metadata[SigMFFile.CAPTURE_KEY] = [{SigMFFile.START_INDEX_KEY: 10}, {SigMFFile.START_INDEX_KEY: 9}]
        with self.assertRaises(ValidationError):
            SigMFFile(self.metadata).validate()

    def test_invalid_annotation_order(self):
        """metadata must have annotations in order"""
        self.metadata[SigMFFile.ANNOTATION_KEY] = [
            {
                SigMFFile.START_INDEX_KEY: 2,
                SigMFFile.LENGTH_INDEX_KEY: 120000,
            },
            {
                SigMFFile.START_INDEX_KEY: 1,
                SigMFFile.LENGTH_INDEX_KEY: 120000,
            },
        ]
        with self.assertRaises(ValidationError):
            SigMFFile(self.metadata).validate()

    def test_annotation_without_sample_count(self):
        """annotation without length should be accepted"""
        self.metadata[SigMFFile.ANNOTATION_KEY] = [{SigMFFile.START_INDEX_KEY: 2}]
        SigMFFile(self.metadata).validate()

    def test_invalid_hash(self):
        """wrong hash raises error on creation"""
        with tempfile.NamedTemporaryFile() as temp_file:
            TEST_FLOAT32_DATA.tofile(temp_file.name)
            self.metadata[SigMFFile.GLOBAL_KEY][SigMFFile.HASH_KEY] = "derp"
            with self.assertRaises(sigmf.error.SigMFFileError):
                SigMFFile(metadata=self.metadata, data_file=temp_file.name)



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Shared test data for tests."""

import numpy as np

from sigmf import SigMFFile, __specification__, __version__

TEST_FLOAT32_DATA = np.arange(16, dtype=np.float32)

TEST_METADATA = {
    SigMFFile.ANNOTATION_KEY: [{SigMFFile.LENGTH_INDEX_KEY: 16, SigMFFile.START_INDEX_KEY: 0}],
    SigMFFile.CAPTURE_KEY: [{SigMFFile.START_INDEX_KEY: 0}],
    SigMFFile.GLOBAL_KEY: {
        SigMFFile.DATATYPE_KEY: "rf32_le",
        SigMFFile.HASH_KEY: "f4984219b318894fa7144519185d1ae81ea721c6113243a52b51e444512a39d74cf41a4cec3c5d000bd7277cc71232c04d7a946717497e18619bdbe94bfeadd6",
        SigMFFile.NUM_CHANNELS_KEY: 1,
        SigMFFile.VERSION_KEY: __specification__,
    },
}

# Data0 is a test of a compliant two capture recording
TEST_U8_DATA0 = list(range(256))
TEST_U8_META0 = {
    SigMFFile.ANNOTATION_KEY: [],
    SigMFFile.CAPTURE_KEY: [
        {SigMFFile.START_INDEX_KEY: 0, SigMFFile.HEADER_BYTES_KEY: 0},
        {SigMFFile.START_INDEX_KEY: 0, SigMFFile.HEADER_BYTES_KEY: 0},
    ],  # very strange..but technically legal?
    SigMFFile.GLOBAL_KEY: {SigMFFile.DATATYPE_KEY: "ru8", SigMFFile.TRAILING_BYTES_KEY: 0},
}
# Data1 is a test of a two capture recording with header_bytes and trailing_bytes set
TEST_U8_DATA1 = [0xFE] * 32 + list(range(192)) + [0xFF] * 32
TEST_U8_META1 = {
    SigMFFile.ANNOTATION_KEY: [],
    SigMFFile.CAPTURE_KEY: [
        {SigMFFile.START_INDEX_KEY: 0, SigMFFile.HEADER_BYTES_KEY: 32},
        {SigMFFile.START_INDEX_KEY: 128},
    ],
    SigMFFile.GLOBAL_KEY: {SigMFFile.DATATYPE_KEY: "ru8", SigMFFile.TRAILING_BYTES_KEY: 32},
}
# Data2 is a test of a two capture recording with multiple header_bytes set
TEST_U8_DATA2 = [0xFE] * 32 + list(range(128)) + [0xFE] * 16 + list(range(128, 192)) + [0xFF] * 16
TEST_U8_META2 = {
    SigMFFile.ANNOTATION_KEY: [],
    SigMFFile.CAPTURE_KEY: [
        {SigMFFile.START_INDEX_KEY: 0, SigMFFile.HEADER_BYTES_KEY: 32},
        {SigMFFile.START_INDEX_KEY: 128, SigMFFile.HEADER_BYTES_KEY: 16},
    ],
    SigMFFile.GLOBAL_KEY: {SigMFFile.DATATYPE_KEY: "ru8", SigMFFile.TRAILING_BYTES_KEY: 16},
}
# Data3 is a test of a three capture recording with multiple header_bytes set
TEST_U8_DATA3 = [0xFE] * 32 + list(range(128)) + [0xFE] * 32 + list(range(128, 192))
TEST_U8_META3 = {
    SigMFFile.ANNOTATION_KEY: [],
    SigMFFile.CAPTURE_KEY: [
        {SigMFFile.START_INDEX_KEY: 0, SigMFFile.HEADER_BYTES_KEY: 32},
        {SigMFFile.START_INDEX_KEY: 32},
        {SigMFFile.START_INDEX_KEY: 128, SigMFFile.HEADER_BYTES_KEY: 32},
    ],
    SigMFFile.GLOBAL_KEY: {SigMFFile.DATATYPE_KEY: "ru8"},
}
# Data4 is a two channel version of Data0
TEST_U8_DATA4 = [0xFE] * 32 + [y for y in list(range(96)) for i in [0, 1]] + [0xFF] * 32
TEST_U8_META4 = {
    SigMFFile.ANNOTATION_KEY: [],
    SigMFFile.CAPTURE_KEY: [
        {SigMFFile.START_INDEX_KEY: 0, SigMFFile.HEADER_BYTES_KEY: 32},
        {SigMFFile.START_INDEX_KEY: 64},
    ],
    SigMFFile.GLOBAL_KEY: {
        SigMFFile.DATATYPE_KEY: "ru8",
        SigMFFile.TRAILING_BYTES_KEY: 32,
        SigMFFile.NUM_CHANNELS_KEY: 2,
    },
}



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later
"""Configuration file for the Sphinx documentation builder."""

import datetime
import re
import sys
from pathlib import Path

# parse info from project files

root = Path(__file__).parent.parent.parent
with open(root / "sigmf" / "__init__.py", "r") as handle:
    init = handle.read()
    toolversion = re.search(r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]', init).group(1)
    specversion = re.search(r'__specification__\s*=\s*[\'"]([^\'"]*)[\'"]', init).group(1)

# autodoc needs special pathing
sys.path.append(str(root))

# -- Project information

project = "sigmf"
author = "Multiple Authors"
copyright = f"2017-{datetime.date.today().year}, {author}"

release = toolversion
version = toolversion

# -- General configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.duration",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",  # allows numpy-style docstrings
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

# -- Options for HTML output

html_theme = "sphinx_rtd_theme"
html_favicon = "https://raw.githubusercontent.com/wiki/sigmf/SigMF/logo/logo-icon-32-folder.png"
html_logo = "https://raw.githubusercontent.com/sigmf/SigMF/refs/heads/main/logo/sigmf_logo.svg"

# -- Options for EPUB output

epub_show_urls = "footnote"

# Method to use variables within rst files
# https://stackoverflow.com/a/69211912/760099

variables_to_export = [
    "toolversion",
    "specversion",
]
frozen_locals = dict(locals())
rst_epilog = '\n'.join(map(lambda x: f".. |{x}| replace:: {frozen_locals[x]}", variables_to_export))
del frozen_locals


# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""SigMFFile Object"""

import codecs
import io
import json
import warnings
from collections import OrderedDict
from pathlib import Path

import numpy as np

from . import __specification__, __version__, schema, sigmf_hash, validate
from .archive import (
    SIGMF_ARCHIVE_EXT,
    SIGMF_COLLECTION_EXT,
    SIGMF_DATASET_EXT,
    SIGMF_METADATA_EXT,
    SigMFArchive,
)
from .error import SigMFAccessError, SigMFError, SigMFFileError
from .utils import dict_merge


class SigMFMetafile:
    VALID_KEYS = {}

    def __init__(self):
        self.version = None
        self.schema = None
        self._metadata = None
        self.shape = None

    def __str__(self):
        return self.dumps()

    def __repr__(self):
        return f"SigMFFile({self})"

    def __iter__(self):
        """special method to iterate through samples"""
        self.iter_position = 0
        return self

    def ordered_metadata(self):
        """
        Get a nicer representation of _metadata. Will sort keys, but put the
        top-level fields 'global', 'captures', 'annotations' in front.

        Returns
        -------
        ordered_meta : OrderedDict
            Cleaner representation of _metadata with top-level keys correctly
            ordered and the rest of the keys sorted.
        """
        ordered_meta = OrderedDict()
        for top_key in self.VALID_KEYS.keys():
            assert top_key in self._metadata
            ordered_meta[top_key] = json.loads(json.dumps(self._metadata[top_key], sort_keys=True))
        # If there are other top-level keys, they go later
        # TODO: sort potential `other` top-level keys
        for oth_key, oth_val in self._metadata.items():
            if oth_key not in self.VALID_KEYS.keys():
                ordered_meta[oth_key] = json.loads(json.dumps(oth_val, sort_keys=True))
        return ordered_meta

    def dump(self, filep, pretty=True):
        """
        Write metadata to a file.

        Parameters
        ----------
        filep : object
            File pointer or something that json.dump() can handle.
        pretty : bool, default True
            When True will write more human-readable output, otherwise will be flat JSON.
        """
        json.dump(
            self.ordered_metadata(),
            filep,
            indent=4 if pretty else None,
            separators=(",", ": ") if pretty else None,
        )

    def dumps(self, pretty=True):
        """
        Get a string representation of the metadata.

        Parameters
        ----------
        pretty : bool, default True
            When True will write more human-readable output, otherwise will be flat JSON.

        Returns
        -------
        string
            String representation of the metadata using json formatter.
        """
        return json.dumps(
            self.ordered_metadata(),
            indent=4 if pretty else None,
            separators=(",", ": ") if pretty else None,
        )


class SigMFFile(SigMFMetafile):
    START_INDEX_KEY = "core:sample_start"
    LENGTH_INDEX_KEY = "core:sample_count"
    GLOBAL_INDEX_KEY = "core:global_index"
    START_OFFSET_KEY = "core:offset"
    NUM_CHANNELS_KEY = "core:num_channels"
    HASH_KEY = "core:sha512"
    VERSION_KEY = "core:version"
    DATATYPE_KEY = "core:datatype"
    FREQUENCY_KEY = "core:frequency"
    HEADER_BYTES_KEY = "core:header_bytes"
    FLO_KEY = "core:freq_lower_edge"
    FHI_KEY = "core:freq_upper_edge"
    SAMPLE_RATE_KEY = "core:sample_rate"
    COMMENT_KEY = "core:comment"
    DESCRIPTION_KEY = "core:description"
    AUTHOR_KEY = "core:author"
    META_DOI_KEY = "core:meta-doi"
    DATA_DOI_KEY = "core:data-doi"
    GENERATOR_KEY = "core:generator"
    LABEL_KEY = "core:label"
    RECORDER_KEY = "core:recorder"
    LICENSE_KEY = "core:license"
    HW_KEY = "core:hw"
    DATASET_KEY = "core:dataset"
    TRAILING_BYTES_KEY = "core:trailing_bytes"
    METADATA_ONLY_KEY = "core:metadata_only"
    EXTENSIONS_KEY = "core:extensions"
    DATETIME_KEY = "core:datetime"
    LAT_KEY = "core:latitude"
    LON_KEY = "core:longitude"
    UUID_KEY = "core:uuid"
    GEOLOCATION_KEY = "core:geolocation"
    COLLECTION_KEY = "core:collection"
    GLOBAL_KEY = "global"
    CAPTURE_KEY = "captures"
    ANNOTATION_KEY = "annotations"
    VALID_GLOBAL_KEYS = [
        AUTHOR_KEY, COLLECTION_KEY, DATASET_KEY, DATATYPE_KEY, DATA_DOI_KEY, DESCRIPTION_KEY, EXTENSIONS_KEY,
        GEOLOCATION_KEY, HASH_KEY, HW_KEY, LICENSE_KEY, META_DOI_KEY, METADATA_ONLY_KEY, NUM_CHANNELS_KEY, RECORDER_KEY,
        SAMPLE_RATE_KEY, START_OFFSET_KEY, TRAILING_BYTES_KEY, VERSION_KEY
    ]
    VALID_CAPTURE_KEYS = [DATETIME_KEY, FREQUENCY_KEY, HEADER_BYTES_KEY, GLOBAL_INDEX_KEY, START_INDEX_KEY]
    VALID_ANNOTATION_KEYS = [
        COMMENT_KEY, FHI_KEY, FLO_KEY, GENERATOR_KEY, LABEL_KEY, LAT_KEY, LENGTH_INDEX_KEY, LON_KEY, START_INDEX_KEY,
        UUID_KEY
    ]
    VALID_KEYS = {GLOBAL_KEY: VALID_GLOBAL_KEYS, CAPTURE_KEY: VALID_CAPTURE_KEYS, ANNOTATION_KEY: VALID_ANNOTATION_KEYS}

    def __init__(self, metadata=None, data_file=None, global_info=None, skip_checksum=False, map_readonly=True):
        """
        API for SigMF I/O

        Parameters
        ----------
        metadata: str or dict, optional
            Metadata for associated dataset.
        data_file: str, optional
            Path to associated dataset.
        global_info: dict, optional
            Set global field shortcut if creating new object.
        skip_checksum: bool, default False
            When True will skip calculating hash on data_file (if present) to check against metadata.
        map_readonly: bool, default True
            Indicates whether assignments on the numpy.memmap are allowed.
        """
        super().__init__()
        self.data_file = None
        self.sample_count = 0
        self._memmap = None
        self.is_complex_data = False  # numpy.iscomplexobj(self._memmap) is not adequate for fixed-point complex case

        self.set_metadata(metadata)
        if global_info is not None:
            self.set_global_info(global_info)
        if data_file is not None:
            self.set_data_file(data_file, skip_checksum=skip_checksum, map_readonly=map_readonly)

    def __len__(self):
        return self._memmap.shape[0]

    def __eq__(self, other):
        """
        Define equality between two `SigMFFile`s.

        Rely on the checksum value in the metadata to decide whether `data_file` is the same since the path of the
        dataset is immaterial to equivalency.
        """
        if isinstance(other, SigMFFile):
            return self._metadata == other._metadata
        return False

    def __next__(self):
        """get next batch of samples"""
        if self.iter_position < len(self):
            # normal batch
            value = self.read_samples(start_index=self.iter_position, count=1)
            self.iter_position += 1
            return value

        else:
            # no more data
            raise StopIteration

    def __getitem__(self, sli):
        mem = self._memmap[sli]  # matches behavior of numpy.ndarray.__getitem__()

        if self._return_type is None:
            return mem

        # is_fixed_point and is_complex
        if self._memmap.ndim == 2:
            # num_channels == 1
            ray = mem[:, 0].astype(self._return_type) + 1.0j * mem[:, 1].astype(self._return_type)
        elif self._memmap.ndim == 3:
            # num_channels > 1
            ray = mem[:, :, 0].astype(self._return_type) + 1.0j * mem[:, :, 1].astype(self._return_type)
        else:
            raise ValueError("unhandled ndim in SigMFFile.__getitem__(); this shouldn't happen")
        return ray[0] if isinstance(sli, int) else ray  # return element instead of 1-element array

    def _get_start_offset(self):
        """
        Return the offset of the first sample.
        """
        return self.get_global_field(self.START_OFFSET_KEY, 0)

    def get_num_channels(self):
        """Returns integer number of channels if present, otherwise 1"""
        return self.get_global_field(self.NUM_CHANNELS_KEY, 1)

    def _is_conforming_dataset(self):
        """
        The dataset is non-conforming if the datafile contains non-sample bytes
        which means global trailing_bytes field is zero or not set, all captures
        `header_bytes` fields are zero or not set. Because we do not necessarily
        know the filename no means of verifying the meta/data filename roots
        match, but this will also check that a data file exists.

        Returns
        -------
        `True` if the dataset is conforming to SigMF, `False` otherwise
        """
        if self.get_global_field(self.TRAILING_BYTES_KEY, 0):
            return False
        for capture in self.get_captures():
            # check for any non-zero `header_bytes` fields in captures segments
            if capture.get(self.HEADER_BYTES_KEY, 0):
                return False
        if self.data_file is not None and not self.data_file.is_file:
            return False
        # if we get here, the file exists and is conforming
        return True

    def get_schema(self):
        """
        Return a schema object valid for the current metadata
        """
        current_metadata_version = self.get_global_info().get(self.VERSION_KEY)
        if self.version != current_metadata_version or self.schema is None:
            self.version = current_metadata_version
            self.schema = schema.get_schema(self.version)
        assert isinstance(self.schema, dict)
        return self.schema

    def set_metadata(self, metadata):
        """
        Read provided metadata as either None (empty), string, bytes, or dictionary.
        """
        if metadata is None:
            # Create empty
            self._metadata = {self.GLOBAL_KEY: {}, self.CAPTURE_KEY: [], self.ANNOTATION_KEY: []}
        elif isinstance(metadata, dict):
            self._metadata = metadata
        elif isinstance(metadata, (str, bytes)):
            self._metadata = json.loads(metadata)
        else:
            raise SigMFError("Unable to interpret provided metadata.")

        # if num_channels missing, default to 1
        if self.get_global_field(self.NUM_CHANNELS_KEY) is None:
            self.set_global_field(self.NUM_CHANNELS_KEY, 1)

        # set version to current implementation
        self.set_global_field(self.VERSION_KEY, __specification__)

    def set_global_info(self, new_global):
        """
        Recursively override existing global metadata with new global metadata.
        """
        self._metadata[self.GLOBAL_KEY] = dict_merge(self._metadata[self.GLOBAL_KEY], new_global)

    def get_global_info(self):
        """
        Returns a dictionary with all the global info.
        """
        try:
            return self._metadata.get(self.GLOBAL_KEY, {})
        except AttributeError:
            return {}

    def set_global_field(self, key, value):
        """
        Inserts a value into the global field.
        """
        self._metadata[self.GLOBAL_KEY][key] = value

    def get_global_field(self, key, default=None):
        """
        Return a field from the global info, or default if the field is not set.
        """
        return self._metadata[self.GLOBAL_KEY].get(key, default)

    def add_capture(self, start_index, metadata=None):
        """
        Insert capture info for sample starting at start_index.
        If there is already capture info for this index, metadata will be merged
        with the existing metadata, overwriting keys if they were previously set.
        """
        assert start_index >= self._get_start_offset()
        capture_list = self._metadata[self.CAPTURE_KEY]
        new_capture = metadata or {}
        new_capture[self.START_INDEX_KEY] = start_index
        # merge if capture exists
        merged = False
        for existing_capture in self._metadata[self.CAPTURE_KEY]:
            if existing_capture[self.START_INDEX_KEY] == start_index:
                existing_capture = dict_merge(existing_capture, new_capture)
                merged = True
        if not merged:
            capture_list += [new_capture]
        # sort captures by start_index
        self._metadata[self.CAPTURE_KEY] = sorted(
            capture_list,
            key=lambda item: item[self.START_INDEX_KEY],
        )

    def get_captures(self):
        """
        Returns a list of dictionaries representing all captures.
        """
        return self._metadata.get(self.CAPTURE_KEY, [])

    def get_capture_info(self, index):
        """
        Returns a dictionary containing all the capture information at sample
        'index'.
        """
        assert index >= self._get_start_offset()
        captures = self._metadata.get(self.CAPTURE_KEY, [])
        assert len(captures) > 0
        cap_info = captures[0]
        for capture in captures:
            if capture[self.START_INDEX_KEY] > index:
                break
            cap_info = capture
        return cap_info

    def get_capture_start(self, index):
        """
        Returns a the start sample index of a given capture, will raise
        SigMFAccessError if this field is missing.
        """
        start = self.get_captures()[index].get(self.START_INDEX_KEY)
        if start is None:
            raise SigMFAccessError("Capture {} does not have required {} key".format(index, self.START_INDEX_KEY))
        return start

    def get_capture_byte_boundarys(self, index):
        """
        Returns a tuple of the file byte range in a dataset of a given SigMF
        capture of the form [start, stop). This function works on either
        compliant or noncompliant SigMF Recordings.
        """
        if index >= len(self.get_captures()):
            raise SigMFAccessError(
                "Invalid captures index {} (only {} captures in Recording)".format(index, len(self.get_captures()))
            )

        start_byte = 0
        prev_start_sample = 0
        for ii, capture in enumerate(self.get_captures()):
            start_byte += capture.get(self.HEADER_BYTES_KEY, 0)
            start_byte += (
                (self.get_capture_start(ii) - prev_start_sample) * self.get_sample_size() * self.get_num_channels()
            )
            prev_start_sample = self.get_capture_start(ii)
            if ii >= index:
                break

        end_byte = start_byte
        if index == len(self.get_captures()) - 1:  # last captures...data is the rest of the file
            end_byte = self.data_file.stat().st_size - self.get_global_field(self.TRAILING_BYTES_KEY, 0)
        else:
            end_byte += (
                (self.get_capture_start(index + 1) - self.get_capture_start(index))
                * self.get_sample_size()
                * self.get_num_channels()
            )
        return (start_byte, end_byte)

    def add_annotation(self, start_index, length=None, metadata=None):
        """
        Insert annotation at start_index with length (if != None).
        """
        assert start_index >= self._get_start_offset()

        new_annot = metadata or {}
        new_annot[self.START_INDEX_KEY] = start_index
        if length is not None:
            assert length >= 1
            new_annot[self.LENGTH_INDEX_KEY] = length

        self._metadata[self.ANNOTATION_KEY] += [new_annot]
        # sort annotations by start_index
        self._metadata[self.ANNOTATION_KEY] = sorted(
            self._metadata[self.ANNOTATION_KEY],
            key=lambda item: item[self.START_INDEX_KEY],
        )

    def get_annotations(self, index=None):
        """
        Get relevant annotations from metadata.

        Parameters
        ----------
        index : int, default None
            If provided returns all annotations that include this sample index.
            When omitted returns all annotations.

        Returns
        -------
        list of dict
            Each dictionary contains one annotation for the sample at `index`.
        """
        annotations = self._metadata.get(self.ANNOTATION_KEY, [])
        if index is None:
            return annotations

        annotations_including_index = []
        for annotation in annotations:
            if index < annotation[self.START_INDEX_KEY]:
                # index is before annotation starts -> skip
                continue
            if self.LENGTH_INDEX_KEY in annotation:
                # Annotation includes sample_count -> check end index
                if index >= annotation[self.START_INDEX_KEY] + annotation[self.LENGTH_INDEX_KEY]:
                    # index is after annotation end -> skip
                    continue

            annotations_including_index.append(annotation)
        return annotations_including_index

    def get_sample_size(self):
        """
        Determines the size of a sample, in bytes, from the datatype of this set.
        For complex data, a 'sample' includes both the real and imaginary part.
        """
        return dtype_info(self.get_global_field(self.DATATYPE_KEY))["sample_size"]

    def _count_samples(self):
        """
        Count, set, and return the total number of samples in the data file.
        If there is no data file but there are annotations, use the sample_count
        from the annotation with the highest end index. If there are no annotations,
        use 0.
        For complex data, a 'sample' includes both the real and imaginary part.
        """
        if self.data_file is None:
            sample_count = self._get_sample_count_from_annotations()
        else:
            header_bytes = sum([c.get(self.HEADER_BYTES_KEY, 0) for c in self.get_captures()])
            file_size = self.data_file.stat().st_size if self.data_size_bytes is None else self.data_size_bytes
            file_data_size = file_size - self.get_global_field(self.TRAILING_BYTES_KEY, 0) - header_bytes  # bytes
            sample_size = self.get_sample_size()  # size of a sample in bytes
            num_channels = self.get_num_channels()
            sample_count = file_data_size // sample_size // num_channels
            if file_data_size % (sample_size * num_channels) != 0:
                warnings.warn(
                    f"File `{self.data_file}` does not contain an integer number of samples across channels. "
                    "It may be invalid data."
                )
            if self._get_sample_count_from_annotations() > sample_count:
                warnings.warn(
                    f"File `{self.data_file}` ends before the final annotation in the corresponding SigMF metadata."
                )
        self.sample_count = sample_count
        return sample_count

    def _get_sample_count_from_annotations(self):
        """
        Returns the number of samples based on annotation with highest end index.
        NOTE: Annotations are ordered by START_INDEX_KEY and not end index, so we
        need to go through all annotations
        """
        annon_sample_count = []
        for annon in self.get_annotations():
            if self.LENGTH_INDEX_KEY in annon:
                # Annotation with sample_count
                annon_sample_count.append(annon[self.START_INDEX_KEY] + annon[self.LENGTH_INDEX_KEY])
            else:
                # Annotation without sample_count - sample count must be at least sample_start
                annon_sample_count.append(annon[self.START_INDEX_KEY])

        if annon_sample_count:
            return max(annon_sample_count)
        else:
            return 0

    def calculate_hash(self):
        """
        Calculates the hash of the data file and adds it to the global section.
        Also returns a string representation of the hash.
        """
        old_hash = self.get_global_field(self.HASH_KEY)
        if self.data_file is not None:
            new_hash = sigmf_hash.calculate_sha512(
                filename=self.data_file,
                offset=self.data_offset,
                size=self.data_size_bytes,
            )
        else:
            new_hash = sigmf_hash.calculate_sha512(
                fileobj=self.data_buffer,
                offset=self.data_offset,
                size=self.data_size_bytes,
            )
        if old_hash is not None:
            if old_hash != new_hash:
                raise SigMFFileError("Calculated file hash does not match associated metadata.")

        self.set_global_field(self.HASH_KEY, new_hash)
        return new_hash

    def set_data_file(
        self, data_file=None, data_buffer=None, skip_checksum=False, offset=0, size_bytes=None, map_readonly=True
    ):
        """
        Set the datafile path, then recalculate sample count. If not skipped,
        update the hash and return the hash string.
        """
        if self.get_global_field(self.DATATYPE_KEY) is None:
            raise SigMFFileError("Error setting data file, the DATATYPE_KEY must be set in the global metadata first.")

        self.data_file = Path(data_file) if data_file else None
        self.data_buffer = data_buffer
        self.data_offset = offset
        self.data_size_bytes = size_bytes
        self._count_samples()

        dtype = dtype_info(self.get_global_field(self.DATATYPE_KEY))
        self.is_complex_data = dtype["is_complex"]
        num_channels = self.get_num_channels()
        self.ndim = 1 if (num_channels < 2) else 2

        complex_int_separates = dtype["is_complex"] and dtype["is_fixedpoint"]
        mapped_dtype_size = dtype["component_size"] if complex_int_separates else dtype["sample_size"]
        mapped_length = None if size_bytes is None else size_bytes // mapped_dtype_size
        mapped_reshape = (-1,)  # we can't use -1 in mapped_length ...
        if num_channels > 1:
            mapped_reshape = mapped_reshape + (num_channels,)
        if complex_int_separates:
            # There is no corresponding numpy type, so we'll have to add another axis, with length of 2
            mapped_reshape = mapped_reshape + (2,)
        self._return_type = dtype["memmap_convert_type"]
        common_args = {"dtype": dtype["memmap_map_type"], "offset": offset}
        try:
            if self.data_file is not None:
                open_mode = "r" if map_readonly else "r+"
                memmap_shape = None if mapped_length is None else (mapped_length,)
                raveled = np.memmap(self.data_file, mode=open_mode, shape=memmap_shape, **common_args)
            elif self.data_buffer is not None:
                buffer_count = -1 if mapped_length is None else mapped_length
                raveled = np.frombuffer(self.data_buffer.getbuffer(), count=buffer_count, **common_args)
            else:
                raise SigMFFileError("In sigmffile.set_data_file(), either data_file or data_buffer must be not None")
        except SigMFFileError:  # TODO include likely exceptions here
            warnings.warn("Failed to create data array from memory-map-file or buffer!")
        else:
            self._memmap = raveled.reshape(mapped_reshape)
            self.shape = self._memmap.shape if (self._return_type is None) else self._memmap.shape[:-1]

        if self.data_file is not None:
            file_name = self.data_file.name
            ext = self.data_file.suffix
            if ext.lower() != SIGMF_DATASET_EXT:
                self.set_global_field(SigMFFile.DATASET_KEY, file_name)

        if skip_checksum:
            return None
        return self.calculate_hash()

    def validate(self):
        """
        Check schema and throw error if issue.
        """
        version = self.get_global_field(self.VERSION_KEY)
        validate.validate(self._metadata, self.get_schema())

    def archive(self, name=None, fileobj=None):
        """Dump contents to SigMF archive format.

        `name` and `fileobj` are passed to SigMFArchive and are defined there.

        """
        archive = SigMFArchive(self, name, fileobj)
        return archive.path

    def tofile(self, file_path, pretty=True, toarchive=False, skip_validate=False):
        """
        Write metadata file or full archive containing metadata & dataset.

        Parameters
        ----------
        file_path : string
            Location to save.
        pretty : bool, default True
            When True will write more human-readable output, otherwise will be flat JSON.
        toarchive : bool, default False
            If True will write both dataset & metadata into SigMF archive format as a single `tar` file.
            If False will only write metadata to `sigmf-meta`.
        """
        if not skip_validate:
            self.validate()
        fns = get_sigmf_filenames(file_path)
        if toarchive:
            self.archive(fns["archive_fn"])
        else:
            with open(fns["meta_fn"], "w") as fp:
                self.dump(fp, pretty=pretty)
                fp.write("\n")  # text files should end in carriage return

    def read_samples_in_capture(self, index=0, autoscale=True):
        """
        Reads samples from the specified captures segment in its entirety.

        Parameters
        ----------
        index : int, default 0
            Captures segment to read samples from.
        autoscale : bool, default True
            If dataset is in a fixed-point representation, scale samples from (min, max) to (-1.0, 1.0)

        Returns
        -------
        data : ndarray
            Samples are returned as an array of float or complex, with number of dimensions equal to NUM_CHANNELS_KEY.
        """
        cb = self.get_capture_byte_boundarys(index)
        if (cb[1] - cb[0]) % (self.get_sample_size() * self.get_num_channels()):
            warnings.warn(
                f"Capture `{index}` in `{self.data_file}` does not contain "
                "an integer number of samples across channels. It may be invalid."
            )

        return self._read_datafile(cb[0], (cb[1] - cb[0]) // self.get_sample_size(), autoscale, False)

    def read_samples(self, start_index=0, count=-1, autoscale=True, raw_components=False):
        """
        Reads the specified number of samples starting at the specified index from the associated data file.

        Parameters
        ----------
        start_index : int, default 0
            Starting sample index from which to read.
        count : int, default -1
            Number of samples to read. -1 will read whole file.
        autoscale : bool, default True
            If dataset is in a fixed-point representation, scale samples from (min, max) to (-1.0, 1.0)
        raw_components : bool, default False
            If True read and return the sample components (individual I & Q for complex, samples for real)
            with no conversions or interleaved channels.

        Returns
        -------
        data : ndarray
            Samples are returned as an array of float or complex, with number of dimensions equal to NUM_CHANNELS_KEY.
        """
        if count == 0:
            raise IOError("Number of samples must be greater than zero, or -1 for all samples.")
        elif start_index + count > self.sample_count:
            raise IOError("Cannot read beyond EOF.")
        if self.data_file is None and not isinstance(self.data_buffer, io.BytesIO):
            if self.get_global_field(self.METADATA_ONLY_KEY, False):
                # only if data_file is `None` allows access to dynamically generated datsets
                raise SigMFFileError("Cannot read samples from a metadata only distribution.")
            else:
                raise SigMFFileError("No signal data file has been associated with the metadata.")
        first_byte = start_index * self.get_sample_size() * self.get_num_channels()

        if not self._is_conforming_dataset():
            warnings.warn(f"Recording dataset appears non-compliant, resulting data may be erroneous")
        return self._read_datafile(first_byte, count * self.get_num_channels(), autoscale, False)

    def _read_datafile(self, first_byte, nitems, autoscale, raw_components):
        """
        internal function for reading samples from datafile
        """
        dtype = dtype_info(self.get_global_field(self.DATATYPE_KEY))
        self.is_complex_data = dtype["is_complex"]
        is_fixedpoint_data = dtype["is_fixedpoint"]
        is_unsigned_data = dtype["is_unsigned"]
        data_type_in = dtype["sample_dtype"]
        component_type_in = dtype["component_dtype"]
        component_size = dtype["component_size"]

        data_type_out = np.dtype("f4") if not self.is_complex_data else np.dtype("f4, f4")
        num_channels = self.get_num_channels()

        if self.data_file is not None:
            fp = open(self.data_file, "rb")
            fp.seek(first_byte, 0)
            data = np.fromfile(fp, dtype=data_type_in, count=nitems)
        elif self.data_buffer is not None:
            data = np.frombuffer(self.data_buffer.getbuffer(), dtype=data_type_in, count=nitems)
        else:
            data = self._memmap

        if num_channels != 1:
            # return reshaped view for num_channels
            # first dimension will be double size if `is_complex_data`
            data = data.reshape(data.shape[0] // num_channels, num_channels)
        if not raw_components:
            data = data.astype(data_type_out)
            if autoscale and is_fixedpoint_data:
                data = data.view(np.dtype("f4"))
                if is_unsigned_data:
                    data -= 2 ** (component_size * 8 - 1)
                data *= 2 ** -(component_size * 8 - 1)
                data = data.view(data_type_out)
            if self.is_complex_data:
                data = data.view(np.complex64)
        else:
            data = data.view(component_type_in)

        if self.data_file is not None:
            fp.close()

        return data


class SigMFCollection(SigMFMetafile):
    VERSION_KEY = "core:version"
    DESCRIPTION_KEY = "core:description"
    AUTHOR_KEY = "core:author"
    COLLECTION_DOI_KEY = "core:collection_doi"
    LICENSE_KEY = "core:license"
    EXTENSIONS_KEY = "core:extensions"
    STREAMS_KEY = "core:streams"
    COLLECTION_KEY = "collection"
    VALID_COLLECTION_KEYS = [
        AUTHOR_KEY,
        COLLECTION_DOI_KEY,
        DESCRIPTION_KEY,
        EXTENSIONS_KEY,
        LICENSE_KEY,
        STREAMS_KEY,
        VERSION_KEY,
    ]
    VALID_KEYS = {COLLECTION_KEY: VALID_COLLECTION_KEYS}

    def __init__(self, metafiles: list = None, metadata: dict = None, base_path=None, skip_checksums: bool = False) -> None:
        """
        Create a SigMF Collection object.

        Parameters
        ----------
        metafiles: list, optional
            A list of SigMF metadata filenames objects comprising the Collection.
            There should be at least one file.
        metadata: dict, optional
            Collection metadata to use, if not provided this will populate a minimal set of default metadata.
            The `core:streams` field will be regenerated automatically.
        base_path : str | bytes | PathLike, optional
            Base path of the collection recordings.
        skip_checksums : bool, optional
            If true will skip calculating checksum on datasets.

        Raises
        ------
        SigMFError
            If metadata files do not exist.
        """
        super().__init__()
        self.skip_checksums = skip_checksums

        if base_path is None:
            self.base_path = Path("")
        else:
            self.base_path = Path(base_path)

        if metadata is None:
            self._metadata = {self.COLLECTION_KEY: {}}
            self._metadata[self.COLLECTION_KEY][self.STREAMS_KEY] = []
        else:
            self._metadata = metadata

        if metafiles is None:
            self.metafiles = []
        else:
            self.set_streams(metafiles)

        # set version to current implementation
        self.set_collection_field(self.VERSION_KEY, __specification__)

        if not self.skip_checksums:
            self.verify_stream_hashes()

    def __len__(self) -> int:
        """
        The length of a collection is the number of streams.
        """
        return len(self.get_stream_names())

    def verify_stream_hashes(self) -> None:
        """
        Compares the stream hashes in the collection metadata to the metadata files.

        Raises
        ------
        SigMFFileError
            If any dataset checksums do not match saved metadata.
        """
        streams = self.get_collection_field(self.STREAMS_KEY, [])
        for stream in streams:
            old_hash = stream.get("hash")
            metafile_name = get_sigmf_filenames(stream.get("name"))["meta_fn"]
            metafile_path = self.base_path / metafile_name
            if Path.is_file(metafile_path):
                new_hash = sigmf_hash.calculate_sha512(filename=metafile_path)
                if old_hash != new_hash:
                    raise SigMFFileError(
                        f"Calculated file hash for {metafile_path} does not match collection metadata."
                    )

    def set_streams(self, metafiles) -> None:
        """
        Configures the collection `core:streams` field from the specified list of metafiles.
        """
        self.metafiles = metafiles
        streams = []
        for metafile in self.metafiles:
            metafile_path = self.base_path / metafile
            if metafile.endswith(".sigmf-meta") and Path.is_file(metafile_path):
                stream = {
                    # name must be string here to be serializable later
                    "name": str(get_sigmf_filenames(metafile)["base_fn"]),
                    "hash": sigmf_hash.calculate_sha512(filename=metafile_path),
                }
                streams.append(stream)
            else:
                raise SigMFFileError(f"Specifed stream file {metafile_path} is not a valid SigMF Metadata file")
        self.set_collection_field(self.STREAMS_KEY, streams)

    def get_stream_names(self) -> list:
        """
        Returns a list of `name` object(s) from the `collection` level `core:streams` metadata.
        """
        return [s.get("name") for s in self.get_collection_field(self.STREAMS_KEY, [])]

    def set_collection_info(self, new_collection: dict) -> None:
        """
        Overwrite the collection info with a new dictionary.
        """
        self._metadata[self.COLLECTION_KEY] = new_collection.copy()

    def get_collection_info(self) -> dict:
        """
        Returns a dictionary with all the collection info.
        """
        try:
            return self._metadata.get(self.COLLECTION_KEY, {})
        except AttributeError:
            return {}

    def set_collection_field(self, key: str, value) -> None:
        """
        Inserts a value into the collection field.
        """
        self._metadata[self.COLLECTION_KEY][key] = value

    def get_collection_field(self, key: str, default=None):
        """
        Return a field from the collection info, or default if the field is not set.
        """
        return self._metadata[self.COLLECTION_KEY].get(key, default)

    def tofile(self, file_path, pretty: bool = True) -> None:
        """
        Write metadata file

        Parameters
        ----------
        file_path : string
            Location to save.
        pretty : bool, default True
            When True will write more human-readable output, otherwise will be flat JSON.
        """
        filenames = get_sigmf_filenames(file_path)
        with open(filenames["collection_fn"], "w") as handle:
            self.dump(handle, pretty=pretty)
            handle.write("\n")  # text files should end in carriage return

    def get_SigMFFile(self, stream_name=None, stream_index=None):
        """
        Returns the SigMFFile instance of the specified stream if it exists
        """
        if stream_name is not None and stream_name not in self.get_stream_names():
            # invalid stream name
            return
        if stream_index is not None and stream_index < len(self):
            stream_name = self.get_stream_names()[stream_index]
        if stream_name is not None:
            metafile = get_sigmf_filenames(stream_name)["meta_fn"]
            metafile_path = self.base_path / metafile
            return fromfile(metafile_path, skip_checksum=self.skip_checksums)


def dtype_info(datatype):
    """
    Parses a datatype string conforming to the SigMF spec and returns a dict
    of values describing the format.

    Keyword arguments:
    datatype -- a SigMF-compliant datatype string
    """
    if datatype is None:
        raise SigMFFileError("Invalid datatype 'None'.")
    output_info = {}
    dtype = datatype.lower()

    is_unsigned_data = "u" in datatype
    is_complex_data = "c" in datatype
    is_fixedpoint_data = "f" not in datatype

    dtype = datatype.lower().split("_")

    byte_order = ""
    if len(dtype) == 2:
        if dtype[1][0] == "l":
            byte_order = "<"
        elif dtype[1][0] == "b":
            byte_order = ">"
        else:
            raise SigMFFileError("Unrecognized endianness specifier: '{}'".format(dtype[1]))
    dtype = dtype[0]
    if "64" in dtype:
        sample_size = 8
    elif "32" in dtype:
        sample_size = 4
    elif "16" in dtype:
        sample_size = 2
    elif "8" in dtype:
        sample_size = 1
    else:
        raise SigMFFileError("Unrecognized datatype: '{}'".format(dtype))
    component_size = sample_size
    if is_complex_data:
        sample_size *= 2
    sample_size = int(sample_size)

    data_type_str = byte_order
    data_type_str += "f" if not is_fixedpoint_data else "u" if is_unsigned_data else "i"
    data_type_str += str(component_size)

    memmap_convert_type = None
    if is_complex_data:
        data_type_str = ",".join((data_type_str, data_type_str))
        memmap_map_type = byte_order
        if is_fixedpoint_data:
            memmap_map_type += ("u" if is_unsigned_data else "i") + str(component_size)
            memmap_convert_type = byte_order + "c8"
        else:
            memmap_map_type += "c" + str(sample_size)
    else:
        memmap_map_type = data_type_str

    data_type_in = np.dtype(data_type_str)
    output_info["sample_dtype"] = data_type_in
    output_info["component_dtype"] = data_type_in["f0"] if is_complex_data else data_type_in
    output_info["sample_size"] = sample_size
    output_info["component_size"] = component_size
    output_info["is_complex"] = is_complex_data
    output_info["is_unsigned"] = is_unsigned_data
    output_info["is_fixedpoint"] = is_fixedpoint_data
    output_info["memmap_map_type"] = memmap_map_type
    output_info["memmap_convert_type"] = memmap_convert_type
    return output_info


def get_dataset_filename_from_metadata(meta_fn, metadata=None):
    """
    Parse provided metadata and return the expected data filename.

    In the case of a metadata-only distribution, or if the file does not exist,
    this will return ``None``.

    Priority for conflicting datasets:

    1. Use the file named ``<stem>.SIGMF_DATASET_EXT`` if it exists.
    2. Use the file in the ``DATASET_KEY`` field (non-compliant dataset) if it exists.
    3. Return ``None`` (may be a metadata-only distribution).
    """
    compliant_filename = get_sigmf_filenames(meta_fn)["data_fn"]
    noncompliant_filename = metadata["global"].get(SigMFFile.DATASET_KEY, None)

    if Path.is_file(compliant_filename):
        if noncompliant_filename:
            warnings.warn(
                f"Compliant Dataset `{compliant_filename}` exists but "
                f"{SigMFFile.DATASET_KEY} is also defined; using `{compliant_filename}`"
            )
        return compliant_filename

    elif noncompliant_filename:
        dir_path = Path(meta_fn).parent
        noncompliant_data_file_path = Path.joinpath(dir_path, noncompliant_filename)
        if Path.is_file(noncompliant_data_file_path):
            if metadata["global"].get(SigMFFile.METADATA_ONLY_KEY, False):
                raise SigMFFileError(
                    f"Schema defines {SigMFFile.DATASET_KEY} "
                    f"but {SigMFFile.METADATA_ONLY_KEY} also exists; using `{noncompliant_filename}`"
                )
            return noncompliant_data_file_path
        else:
            raise SigMFFileError(
                f"Non-Compliant Dataset `{noncompliant_filename}` is specified in {SigMFFile.DATASET_KEY} "
                "but does not exist!"
            )
    return None


def fromarchive(archive_path, dir=None, skip_checksum=False):
    """Extract an archive and return a SigMFFile.

    The `dir` parameter is no longer used as this function has been changed to
    access SigMF archives without extracting them.
    """
    from .archivereader import SigMFArchiveReader
    return SigMFArchiveReader(archive_path, skip_checksum=skip_checksum).sigmffile


def fromfile(filename, skip_checksum=False):
    """
    Creates and returns a SigMFFile or SigMFCollection instance with metadata
    loaded from the specified file. The filename may be that of either a
    sigmf-meta file, a sigmf-data file, a sigmf-collection file, or a sigmf
    archive.

    Parameters
    ----------
    filename: str | bytes | PathLike
        Path for SigMF Metadata, Dataset, Archive or Collection (with or without extension).
    skip_checksum: bool, default False
        When True will not read entire dataset to calculate hash.

    Returns
    -------
    object
        SigMFFile with dataset & metadata or a SigMFCollection depending on file type.
    """
    fns = get_sigmf_filenames(filename)
    meta_fn = fns["meta_fn"]
    archive_fn = fns["archive_fn"]
    collection_fn = fns["collection_fn"]

    # extract the extension to check whether we are dealing with an archive, collection, etc.
    file_path = Path(filename)
    ext = file_path.suffix

    if (ext.lower().endswith(SIGMF_ARCHIVE_EXT) or not Path.is_file(meta_fn)) and Path.is_file(archive_fn):
        return fromarchive(archive_fn, skip_checksum=skip_checksum)

    if (ext.lower().endswith(SIGMF_COLLECTION_EXT) or not Path.is_file(meta_fn)) and Path.is_file(collection_fn):
        collection_fp = open(collection_fn, "rb")
        bytestream_reader = codecs.getreader("utf-8")
        mdfile_reader = bytestream_reader(collection_fp)
        metadata = json.load(mdfile_reader)
        collection_fp.close()

        dir_path = meta_fn.parent
        return SigMFCollection(metadata=metadata, base_path=dir_path, skip_checksums=skip_checksum)

    else:
        meta_fp = open(meta_fn, "rb")
        bytestream_reader = codecs.getreader("utf-8")
        mdfile_reader = bytestream_reader(meta_fp)
        metadata = json.load(mdfile_reader)
        meta_fp.close()

        data_fn = get_dataset_filename_from_metadata(meta_fn, metadata)
        return SigMFFile(metadata=metadata, data_file=data_fn, skip_checksum=skip_checksum)


def get_sigmf_filenames(filename):
    """
    Safely returns a set of SigMF file paths given an input filename.

    Parameters
    ----------
    filename : str | bytes | PathLike
        The SigMF filename with any extension.

    Returns
    -------
    dict with 'data_fn', 'meta_fn', and 'archive_fn' as keys.
    """
    stem_path = Path(filename)
    # If the path has a sigmf suffix, remove it. Otherwise do not remove the
    # suffix, because the filename might contain '.' characters which are part
    # of the filename rather than an extension.
    sigmf_suffixes = [
        SIGMF_DATASET_EXT, SIGMF_METADATA_EXT,
        SIGMF_ARCHIVE_EXT, SIGMF_COLLECTION_EXT,
    ]
    if stem_path.suffix in sigmf_suffixes:
        with_suffix_path = stem_path
        stem_path = stem_path.with_suffix("")
    else:
        # Add a dummy suffix to prevent the .with_suffix() calls below from
        # overriding part of the filename which is interpreted as a suffix
        with_suffix_path = stem_path.with_name(f"{stem_path.name}{SIGMF_DATASET_EXT}")

    return {
        "base_fn": stem_path,
        "data_fn": with_suffix_path.with_suffix(SIGMF_DATASET_EXT),
        "meta_fn": with_suffix_path.with_suffix(SIGMF_METADATA_EXT),
        "archive_fn": with_suffix_path.with_suffix(SIGMF_ARCHIVE_EXT),
        "collection_fn": with_suffix_path.with_suffix(SIGMF_COLLECTION_EXT),
    }



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Utilities"""

import re
import sys
from copy import deepcopy
from datetime import datetime, timezone

import numpy as np

from .error import SigMFError

SIGMF_DATETIME_ISO8601_FMT = "%Y-%m-%dT%H:%M:%S.%fZ"


def get_sigmf_iso8601_datetime_now() -> str:
    """Get current UTC time as iso8601 string."""
    return datetime.now(timezone.utc).strftime(SIGMF_DATETIME_ISO8601_FMT)


def parse_iso8601_datetime(string: str) -> datetime:
    """
    Parse an iso8601 string as a datetime struct.
    Input string (indicated by final Z) is in UTC tz.

    Example
    -------
    >>> parse_iso8601_datetime("1955-11-05T06:15:00Z")
    datetime.datetime(1955, 11, 5, 6, 15, tzinfo=datetime.timezone.utc)
    """
    match = re.match(r"^(?P<dt>.*)(?P<frac>\.[0-9]{7,})Z$", string)
    if match:
        # string exceeds max precision allowed by strptime -> truncate to µs
        groups = match.groupdict()
        length = min(7, len(groups["frac"]))
        string = "".join([groups["dt"], groups["frac"][:length], "Z"])

    if "." in string:
        # parse float seconds
        format_str = SIGMF_DATETIME_ISO8601_FMT
    else:
        # parse whole seconds
        format_str = SIGMF_DATETIME_ISO8601_FMT.replace(".%f", "")
    return datetime.strptime(string, format_str).replace(tzinfo=timezone.utc)


def dict_merge(a_dict: dict, b_dict: dict) -> dict:
    """
    Recursively merge `b_dict` into `a_dict`.
    `b_dict[key]` will overwrite `a_dict[key]` if it exists.

    Example
    -------
    >>> a, b = {0:0, 1:2}, {1:3, 2:4}
    >>> dict_merge(a, b)
    {0: 0, 1: 3, 2: 4}
    """
    if not isinstance(b_dict, dict):
        return b_dict
    result = deepcopy(a_dict)
    for key, value in b_dict.items():
        if key in result and isinstance(result[key], dict):
            result[key] = dict_merge(result[key], value)
        else:
            result[key] = deepcopy(value)
    return result


def get_endian_str(ray: np.ndarray) -> str:
    """Return SigMF compatible endianness string for a numpy array"""
    if not isinstance(ray, np.ndarray):
        raise SigMFError("Argument must be a numpy array")
    atype = ray.dtype

    if atype.byteorder == "<":
        return "_le"
    if atype.byteorder == ">":
        return "_be"
    # endianness is then either '=' (native) or '|' (doesn't matter)
    return "_le" if sys.byteorder == "little" else "_be"


def get_data_type_str(ray: np.ndarray) -> str:
    """
    Return the SigMF datatype string for the datatype of numpy array `ray`.

    NOTE: this function only supports native numpy types so interleaved complex
    integer types are not supported.
    """
    if not isinstance(ray, np.ndarray):
        raise SigMFError("Argument must be a numpy array")
    atype = ray.dtype
    if atype.kind not in ("u", "i", "f", "c"):
        raise SigMFError("Unsupported data type:", atype)
    data_type_str = ""
    if atype.kind == "c":
        data_type_str += "cf"
        # units are component bits, numpy complex types len(I)+len(Q)
        data_type_str += str(atype.itemsize * 8 // 2)
    elif atype.kind == "f":
        data_type_str += "rf"
        data_type_str += str(atype.itemsize * 8)  # itemsize in bits
    elif atype.kind in ("u", "i"):
        data_type_str += "r" + atype.kind
        data_type_str += str(atype.itemsize * 8)  # itemsize in bits
    if atype.itemsize > 1:
        # only append endianness for types over 8 bits
        data_type_str += get_endian_str(ray)
    return data_type_str



# Copyright: Multiple Authors
#
# This file is part of sigmf-python. https://github.com/sigmf/sigmf-python
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Tests for SigMFArchiveReader"""

import unittest
from tempfile import NamedTemporaryFile

import numpy as np

import sigmf
from sigmf import SigMFArchiveReader, SigMFFile, __specification__


class TestArchiveReader(unittest.TestCase):
    def setUp(self):
        # in order to check shapes we need some positive number of samples to work with
        # number of samples should be lowest common factor of num_channels
        self.raw_count = 16
        self.lut = {
            "i8": np.int8,
            "u8": np.uint8,
            "i16": np.int16,
            "u16": np.uint16,
            "u32": np.uint32,
            "i32": np.int32,
            "f32": np.float32,
            "f64": np.float64,
        }

    def test_access_data_without_untar(self):
        """iterate through datatypes and verify IO is correct"""
        temp_data = NamedTemporaryFile()
        temp_archive = NamedTemporaryFile(suffix=".sigmf")

        for key, dtype in self.lut.items():
            # for each type of storage
            temp_samples = np.arange(self.raw_count, dtype=dtype)
            temp_samples.tofile(temp_data.name)
            for num_channels in [1, 4, 8]:
                # for single or 8 channel
                for complex_prefix in ["r", "c"]:
                    # for real or complex
                    target_count = self.raw_count
                    temp_meta = SigMFFile(
                        data_file=temp_data.name,
                        global_info={
                            SigMFFile.DATATYPE_KEY: f"{complex_prefix}{key}_le",
                            SigMFFile.NUM_CHANNELS_KEY: num_channels,
                        },
                    )
                    temp_meta.tofile(temp_archive.name, toarchive=True)

                    readback = SigMFArchiveReader(temp_archive.name)
                    readback_samples = readback[:]

                    if complex_prefix == "c":
                        # complex data will be half as long
                        target_count //= 2
                        self.assertTrue(np.all(np.iscomplex(readback_samples)))
                    if num_channels != 1:
                        # check expected # of channels
                        self.assertEqual(
                            readback_samples.ndim,
                            2,
                            "Mismatch in shape of readback samples.",
                        )
                    target_count //= num_channels

                    self.assertEqual(
                        target_count,
                        temp_meta._count_samples(),
                        "Mismatch in expected metadata length.",
                    )
                    self.assertEqual(
                        target_count,
                        len(readback),
                        "Mismatch in expected readback length",
                    )


def test_archiveread_data_file_unchanged(test_sigmffile):
    with NamedTemporaryFile(suffix=".sigmf") as temp_file:
        input_samples = test_sigmffile.read_samples()
        test_sigmffile.archive(temp_file.name)
        arc = sigmf.sigmffile.fromfile(temp_file.name)
        output_samples = arc.read_samples()

        assert np.array_equal(input_samples, output_samples)
